import numpy as np
import matplotlib.pyplot as plt
from poisson_approval.utils.Util import initialize_random_seeds
from poisson_approval.profiles.ProfileOrdinal import ProfileOrdinal
from poisson_approval.generators.GeneratorProfileOrdinalUniform import GeneratorProfileOrdinalUniform


# noinspection PyUnresolvedReferences
class NiceStatsProfileOrdinal:
    """Compute nice stats on ordinal profiles.

    Parameters
    ----------
    tests_profile : list
        A list of pairs ``(test, name)``. ``test`` is a function ``ProfileOrdinal -> bool``. ``name`` is a
        string.
    tests_strategy : list
        A list of pairs ``(test, name)``. ``test`` is a function ``StrategyOrdinal -> bool``. ``name`` is a string.
        For these tests, we compute only the probability that a equilibrium meeting the test exists.
    tests_strategy_dist : list
        A list of pairs ``(test, name)``. ``test`` is a function ``StrategyOrdinal -> bool``. ``name`` is a string.
        For these tests, we compute the distribution of numbers of equilibria meeting the test.
    tests_strategy_winners : list
        A list of pairs ``(test, name)``. ``test`` is a function ``StrategyOrdinal -> bool``. ``name`` is a string.
        For these tests, we compute the distribution of numbers of winners in equilibria meeting the test.
    conditional_on : callable
        A function ``ProfileOrdinal -> bool``.
    generator_profiles : callable
        A callable that inputs nothing and outputs a profile. Default: ``GeneratorProfileOrdinalUniform()``.

    Notes
    -----
    The tests in `tests_profile` concern the profile: what proportions of profiles meet the condition?

    The tests in `tests_strategy` concern the strategies, in the cases where they are equilibrium. For each
    ordinal profile, we compute the probability (for a uniform distribution of utilities) that at least one strategy
    is an equilibrium and meet the given condition. This is used to draw plots `à la Antonin`.

    The tests in `tests_strategy_dist` or `tests_strategy_winners` concern also the strategies, in the cases where
    they are equilibrium.

    Examples
    --------
        >>> initialize_random_seeds()
        >>> nice_stats = NiceStatsProfileOrdinal(
        ...     tests_profile=[
        ...         (lambda profile: any([strategy for strategy in profile.analyzed_strategies.equilibria
        ...                               if strategy.profile.condorcet_winners == strategy.winners]),
        ...          'There exists a true equilibrium electing the CW'),
        ...     ],
        ...     tests_strategy=[
        ...         (lambda strategy: strategy.profile.condorcet_winners == strategy.winners,
        ...          'There exists an equilibrium that elects the CW')
        ...     ],
        ...     tests_strategy_dist=[
        ...         (lambda strategy: strategy.profile.condorcet_winners == strategy.winners,
        ...          'There exists an equilibrium that elects the CW')
        ...     ],
        ...     tests_strategy_winners = [
        ...         (lambda sigma: True, 'Number of possible winners')
        ...     ],
        ...     conditional_on=lambda profile: profile.is_profile_condorcet == 1.
        ... )
        >>> nice_stats.run(n_samples=10)
        >>> profile = nice_stats.find_example('There exists a true equilibrium electing the CW', False)
        >>> print(profile)
        <abc: 0.4236547993389047, acb: 0.12122838365799216, bac: 0.0039303209304278885, bca: 0.05394987214431912, \
cab: 0.1124259903007756, cba: 0.2848106336275805> (Condorcet winner: a)
    """

    def __init__(self, tests_profile=None, tests_strategy=None, tests_strategy_dist=None, tests_strategy_winners=None,
                 conditional_on=None, generator_profiles=None):
        self.tests_profile = [] if tests_profile is None else tests_profile
        self.tests_strategy = [] if tests_strategy is None else tests_strategy
        self.tests_strategy_dist = [] if tests_strategy_dist is None else tests_strategy_dist
        self.tests_strategy_winners = [] if tests_strategy_winners is None else tests_strategy_winners
        self.conditional_on = (lambda profile: True) if conditional_on is None else conditional_on
        self.generator_profiles = GeneratorProfileOrdinalUniform() if generator_profiles is None else generator_profiles
        # Computed variables
        self.n_samples = None
        self.profiles = None
        self.results_profile = None
        self.results_strategy = None
        self.results_strategy_dist = None
        self.results_strategy_winners = None

    def run(self, n_samples):
        """Run the simulations and store the results.

        Parameters
        ----------
        n_samples : int
            Number of profiles meeting the precondition `conditional_on`.
        """
        self.n_samples = n_samples
        i_sample = 0
        self.profiles = []
        self.results_profile = [[] for _ in range(len(self.tests_profile))]
        self.results_strategy = [[] for _ in range(len(self.tests_strategy))]
        self.results_strategy_dist = [np.zeros(2 ** 6) for _ in range(len(self.tests_strategy_dist))]
        self.results_strategy_winners = [np.zeros(4) for _ in range(len(self.tests_strategy_winners))]
        while i_sample < n_samples:
            profile = self.generator_profiles()
            if self.conditional_on(profile):
                i_sample += 1
            else:
                continue  # pragma: no cover - I do not understand why it pretends not to be covered, whereas it is...
            self.profiles.append(profile.d_ranking_share)
            for i, (test, _) in enumerate(self.tests_profile):
                self.results_profile[i].append(test(profile))
            for i, (test, _) in enumerate(self.tests_strategy):
                self.results_strategy[i].append(profile.proba_equilibrium(test=test))
            for i, (test, _) in enumerate(self.tests_strategy_dist):
                histogram = profile.distribution_equilibria(test=test)
                self.results_strategy_dist[i][:len(histogram)] += histogram
            for i, (test, _) in enumerate(self.tests_strategy_winners):
                histogram = profile.distribution_winners(test=test)
                self.results_strategy_winners[i][:len(histogram)] += histogram
        for histogram in self.results_strategy_dist:
            histogram /= n_samples
        for histogram in self.results_strategy_winners:
            histogram /= n_samples

    def plot_test_strategy(self, test, ylabel=True, legend=False, replacement_name=None, style=''):  # pragma: no cover
        """Plot a test on strategy.

        Parameters
        ----------
        test : str or int
            Name or index of a test in `tests_strategy`.
        ylabel : bool
            If True, the name is used in the y-label. Otherwise, the label is `P` (probability).
        legend : bool
            If True, the legend is displayed.
        replacement_name : str, optional
            If specified, it will be used instead of the test name in the plot.
        style : str
            Cf. :meth:`matplotlib.pyplot.plot`.
        """
        if isinstance(test, str):
            i = [name for _, name in self.tests_strategy].index(test)
        else:
            i = test
        (_, name) = self.tests_strategy[i]
        if replacement_name is not None:
            name = replacement_name
        plt.plot(np.arange(0, 1, 1 / self.n_samples), sorted(self.results_strategy[i]), style, label=name)
        plt.ylim(-0.05, 1.05)
        plt.xlabel('Cumulative share of ordinal preference profiles')
        if ylabel:
            plt.ylabel('$\mathbb{P}$(%s)' % name)
        else:
            plt.ylabel('$\mathbb{P}$')
        if legend:
            plt.legend()

    def plot_cutoff(self, test, left='', right='', style=''):  # pragma: no cover
        """Plot the cutoff of a test on the profile.

        Parameters
        ----------
        test : str or int
            Name or index of a test in `tests_profile`.
        left : str
            Text to be written on the left.
        right : str
            Text to be written on the right.
        style : str
            Cf. :meth:`matplotlib.pyplot.plot`.
        """
        if isinstance(test, str):
            i = [name for _, name in self.tests_profile].index(test)
        else:
            i = test
        (_, name) = self.tests_profile[i]
        x = np.sum(self.results_profile[i]) / self.n_samples
        plt.plot([1 - x, 1 - x], [0., 1.], style)
        plt.text((1 - x) / 2, 0.1, left, horizontalalignment='center', verticalalignment='center')
        plt.text(1 - x / 2, 0.1, right, horizontalalignment='center', verticalalignment='center')

    def display_results(self):  # pragma: no cover
        """Display the results."""
        for i, (_, name) in enumerate(self.tests_profile):
            print('P(%s) = %s' % (name, np.sum(self.results_profile[i]) / self.n_samples))
        for i, (_, name) in enumerate(self.tests_strategy):
            self.plot_test_strategy(test=i)
            plt.show()
        for i, (_, name) in enumerate(self.tests_strategy_dist):
            print('P(%s) :' % name)
            true_dim = np.max(np.flatnonzero(self.results_strategy_dist[i])) + 1
            print({k: self.results_strategy_dist[i][k] for k in range(true_dim)})
            plt.bar(range(true_dim), self.results_strategy_dist[i][:true_dim])
            plt.xticks(range(true_dim))
            plt.xlabel(name)
            plt.ylabel('$\mathbb{P}$')
            plt.ylim(0, 1)
            plt.show()
        for i, (_, name) in enumerate(self.tests_strategy_winners):
            print('%s :' % name)
            true_dim = 4
            print({k: self.results_strategy_winners[i][k] for k in range(true_dim)})
            plt.bar(range(true_dim), self.results_strategy_winners[i][:true_dim])
            plt.xticks(range(true_dim))
            plt.xlabel(name)
            plt.ylabel('$\mathbb{P}$')
            plt.ylim(0, 1)
            plt.show()

    def find_example(self, test, value=True):
        """Find an example profile.

        Parameters
        ----------
        test : str or int
            Name or index of a test in `tests_profile`.
        value : bool
            Whether we want an example (True) or a counter-example (False).

        Returns
        -------
        ProfileOrdinal
            A profile for which the test returned `value` in the simulation.
        """
        if isinstance(test, str):
            i_test = [name for _, name in self.tests_profile].index(test)
        else:
            i_test = test
        i_profile = self.results_profile[i_test].index(value)
        return ProfileOrdinal(self.profiles[i_profile])



import warnings
from math import isclose
from poisson_approval.constants.constants import *
from poisson_approval.utils.Util import ballot_one, ballot_two, ballot_one_two, ballot_one_three, sort_ballot, \
    product_dict, ballot_low_u, ballot_high_u
from poisson_approval.utils.SetPrintingInOrder import SetPrintingInOrder
from poisson_approval.profiles.ProfileCardinal import ProfileCardinal
from poisson_approval.tau_vector.TauVector import TauVector
from poisson_approval.strategies.StrategyTwelve import StrategyTwelve
from poisson_approval.containers.AnalyzedStrategies import AnalyzedStrategies
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros
from poisson_approval.constants.EquilibriumStatus import EquilibriumStatus
from poisson_approval.utils.UtilCache import cached_property


# noinspection PyUnresolvedReferences
class ProfileTwelve(ProfileCardinal):
    """A profile of preference with twelve types.

    Parameters
    ----------
    d_type_share : dict
        E.g. ``{'ab_c': 0.4, 'c_ab': 0.6}``. ``d_type_share['ab_c']`` is the probability
        that a voter prefers candidate ``a``, then candidate ``b``, then candidate ``c``, with a utility for ``b``
        that is infinitely close to 1. In contrast, ``d_type_share['a_bc']`` is the probability that a voter prefers
        ``a`` then ``b`` then ``c``, with a utility for ``b`` that is infinitely close to 0.
    normalization_warning : bool
        Whether a warning should be issued if the input distribution is not normalized.
    ratio_sincere : Number
        The ratio of sincere voters, in the interval [0, 1]. This is used for :meth:`tau`.
    ratio_fanatic : Number
        The ratio of fanatic voters, in the interval [0, 1]. This is used for :meth:`tau`. The sum of `ratio_sincere`
        and `ratio_fanatic` must not exceed 1.
    voting_rule : str
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``.

    Notes
    -----
    If the input distribution `d_type_share` is not normalized, the profile will be normalized anyway and a
    warning is issued (unless `normalization_warning` is False).

    Examples
    --------
        >>> from fractions import Fraction
        >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
        ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
        >>> profile
        ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(3, 5), 'c_ab': Fraction(1, 5), \
'ca_b': Fraction(1, 10)})
        >>> print(profile)
        <ab_c: 1/10, b_ac: 3/5, c_ab: 1/5, ca_b: 1/10> (Condorcet winner: b)
        >>> profile.c_ab
        Fraction(1, 5)
        >>> profile.d_type_share['c_ab']  # Alternate syntax for profile.c_ab
        Fraction(1, 5)
        >>> profile.cab
        Fraction(3, 10)
        >>> profile.d_ranking_share['cab']  # Alternate syntax for profile.cab
        Fraction(3, 10)
        >>> profile.weighted_maj_graph
        array([[0, Fraction(-1, 5), Fraction(2, 5)],
               [Fraction(1, 5), 0, Fraction(2, 5)],
               [Fraction(-2, 5), Fraction(-2, 5), 0]], dtype=object)
        >>> profile.condorcet_winners
        Winners({'b'})
        >>> profile.is_profile_condorcet
        1.0
        >>> profile.has_majority_favorite  # Is one candidate 'top' in a majority of ballots?
        True
        >>> profile.has_majority_ranking  # Does one ranking represent a majority of ballots?
        True
        >>> profile.is_single_peaked  # Is the profile single-peaked?
        True
        >>> profile.support_in_rankings
        {'abc', 'bac', 'cab'}
        >>> profile.is_generic_in_rankings  # Are all rankings there?
        False

    In the following example, one third of the voters are sincere:

        >>> from fractions import Fraction
        >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
        ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)},
        ...                         ratio_sincere=Fraction(1, 3))
        >>> tau_sincere = profile.tau_sincere
        >>> print(tau_sincere)
        <ab: 1/10, ac: 1/10, b: 3/5, c: 1/5> ==> b
        >>> strategy = StrategyTwelve({'abc': 'a', 'bac': 'b', 'cab': 'utility-dependent'})
        >>> tau_strategic = profile.tau_strategic(strategy)
        >>> print(tau_strategic)
        <a: 1/10, ac: 1/10, b: 3/5, c: 1/5> ==> b
        >>> tau = profile.tau(strategy)
        >>> print(tau)
        <a: 1/15, ab: 1/30, ac: 1/10, b: 3/5, c: 1/5> ==> b
    """

    def __init__(self, d_type_share, normalization_warning=True, ratio_sincere=0, ratio_fanatic=0,
                 voting_rule=APPROVAL):
        super().__init__(ratio_sincere=ratio_sincere, ratio_fanatic=ratio_fanatic, voting_rule=voting_rule)
        # Populate the dictionary
        self.d_type_share = DictPrintingInOrderIgnoringZeros({t: 0 for t in TWELVE_TYPES})
        for t, share in d_type_share.items():
            self.d_type_share[t] += share
        # Normalize if necessary
        total = sum(self.d_type_share.values())
        if not isclose(total, 1.):
            if normalization_warning:
                warnings.warn("Warning: profile is not normalized, I will normalize it.")
            for t in self.d_type_share.keys():
                self.d_type_share[t] /= total

    def have_ranking_with_utility_above_u(self, ranking, u):
        """Share of voters who have a given ranking and strictly above a given utility for their middle candidate.

        Cf. :meth:`ProfileCardinal.have_ranking_with_utility_above_u`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> profile.have_ranking_with_utility_above_u(ranking='cab', u=.5)
            Fraction(1, 10)
            >>> profile.have_ranking_with_utility_above_u(ranking='cab', u=0)
            Fraction(3, 10)
            >>> profile.have_ranking_with_utility_above_u(ranking='cab', u=1)
            0
        """
        high_u = self.d_type_share[ranking[:2] + '_' + ranking[2:]]  # E.g. ab_c
        low_u = self.d_type_share[ranking[:1] + '_' + ranking[1:]]   # E.g. a_bc
        if u == 1:
            return 0
        if u == 0:
            return high_u + low_u
        return high_u

    def have_ranking_with_utility_u(self, ranking, u):
        """Share of voters who have a given ranking and a given utility for their middle candidate.

        Cf. :meth:`ProfileCardinal.have_ranking_with_utility_u`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> profile.have_ranking_with_utility_u(ranking='cab', u=.5)
            0
        """
        return 0

    def have_ranking_with_utility_below_u(self, ranking, u):
        """Share of voters who have a given ranking and strictly below a given utility for their middle candidate.

        Cf. :meth:`ProfileCardinal.have_ranking_with_utility_below_u`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> profile.have_ranking_with_utility_below_u(ranking='cab', u=.5)
            Fraction(1, 5)
            >>> profile.have_ranking_with_utility_below_u(ranking='cab', u=1)
            Fraction(3, 10)
            >>> profile.have_ranking_with_utility_below_u(ranking='cab', u=0)
            0
        """
        high_u = self.d_type_share[ranking[:2] + '_' + ranking[2:]]  # E.g. ab_c
        low_u = self.d_type_share[ranking[:1] + '_' + ranking[1:]]   # E.g. a_bc
        if u == 1:
            return high_u + low_u
        if u == 0:
            return 0
        return low_u

    def __repr__(self):
        """
        >>> from fractions import Fraction
        >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
        ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)},
        ...                         ratio_sincere=Fraction(1, 10), ratio_fanatic=Fraction(1, 5))
        >>> profile
        ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(3, 5), 'c_ab': Fraction(1, 5), \
'ca_b': Fraction(1, 10)}, ratio_sincere=Fraction(1, 10), ratio_fanatic=Fraction(1, 5))
        """
        arguments = repr(self.d_type_share)
        if self.ratio_sincere > 0:
            arguments += ', ratio_sincere=%r' % self.ratio_sincere
        if self.ratio_fanatic > 0:
            arguments += ', ratio_fanatic=%r' % self.ratio_fanatic
        if self.voting_rule != APPROVAL:
            arguments += ', voting_rule=%r' % self.voting_rule
        return 'ProfileTwelve(%s)' % arguments

    def __str__(self):
        """
        >>> from fractions import Fraction
        >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
        ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)},
        ...                         ratio_sincere=Fraction(1, 10), ratio_fanatic=Fraction(1, 5))
        >>> print(profile)
        <ab_c: 1/10, b_ac: 3/5, c_ab: 1/5, ca_b: 1/10> (Condorcet winner: b) (ratio_sincere: 1/10) (ratio_fanatic: 1/5)
        """
        result = '<%s>' % str(self.d_type_share)[1:-1]
        if self.is_profile_condorcet:
            result += ' (Condorcet winner: %s)' % self.condorcet_winners
        if self.ratio_sincere > 0:
            result += ' (ratio_sincere: %s)' % self.ratio_sincere
        if self.ratio_fanatic > 0:
            result += ' (ratio_fanatic: %s)' % self.ratio_fanatic
        if self.voting_rule != APPROVAL:
            result += ' (%s)' % self.voting_rule
        return result

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')

    def __eq__(self, other):
        """Equality test.

        Parameters
        ----------
        other : Object

        Returns
        -------
        bool
            True iff this profile is equal to `other`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> profile == ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                           'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            True
        """
        return (isinstance(other, ProfileTwelve)
                and self.d_type_share == other.d_type_share
                and self.ratio_sincere == other.ratio_sincere
                and self.ratio_fanatic == other.ratio_fanatic
                and self.voting_rule == other.voting_rule)

    # Standardized version of the profile (makes it unique, up to permutations)

    @cached_property
    def standardized_version(self):
        """ProfileTwelve : Standardized version of the profile (makes it unique, up to permutations of the candidates).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> print(profile.standardized_version)
            <a_bc: 3/5, ba_c: 1/10, c_ba: 1/5, cb_a: 1/10> (Condorcet winner: a)
            >>> profile.is_standardized
            False
        """
        def translate(s, permute):
            return s.replace('a', permute[0]).replace('b', permute[1]).replace('c', permute[2])

        best_d = {}
        best_signature = []
        for perm in XYZ_PERMUTATIONS:
            d_test = {translate(t, perm): share for t, share in self.d_type_share.items()}
            signature_test = [d_test[t] for t in XYZ_TWELVE_TYPES]
            if signature_test > best_signature:
                best_signature = signature_test
                best_d = d_test
        return ProfileTwelve({t: best_d[xyz_t] for t, xyz_t in zip(TWELVE_TYPES, XYZ_TWELVE_TYPES)},
                             ratio_sincere=self.ratio_sincere, ratio_fanatic=self.ratio_fanatic,
                             voting_rule=self.voting_rule)

    @cached_property
    def has_majority_type(self):
        """bool : Whether there is a majority type (a type shared by strictly more than half of the voters).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> profile.has_majority_type
            True
        """
        return max(self.d_type_share.values()) > 0.5

    # Has full support
    @cached_property
    def support_in_types(self):
        """:class:`SetPrintingInOrder` of str : Support of the profile (in terms of types).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> profile.support_in_types
            {'ab_c', 'b_ac', 'c_ab', 'ca_b'}
        """
        return SetPrintingInOrder({t for t, share in self.d_type_share.items() if share > 0})

    @cached_property
    def is_generic_in_types(self):
        """bool : Whether the profile is generic in types (contains all types).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> profile.is_generic_in_types
            False
        """
        return 0 not in self.d_type_share.values()

    # Tau and strategy-related stuff

    def tau_strategic(self, strategy):
        """Tau-vector associated to a strategy.

        Parameters
        ----------
        strategy : StrategyTwelve
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        TauVector
            Tau-vector associated to this profile and strategy `strategy`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> strategy = StrategyTwelve({'abc': 'ab', 'bac': 'b', 'cab': 'utility-dependent'})
            >>> tau_strategic = profile.tau_strategic(strategy)
            >>> print(tau_strategic)
            <ab: 1/10, ac: 1/10, b: 3/5, c: 1/5> ==> b
        """
        assert self.voting_rule == strategy.voting_rule
        t = {ballot: 0 for ballot in BALLOTS_WITHOUT_INVERSIONS}
        for ranking, ballot in strategy.d_ranking_ballot.items():
            if self.d_ranking_share[ranking] == 0:
                continue
            # For a ranking abc, ballot can be real ballots (e.g. 'a', 'ab'), '' or 'utility-dependent'.
            if ballot == UTILITY_DEPENDENT:
                t[ballot_low_u(ranking, self.voting_rule)] += self.have_ranking_with_utility_below_u(ranking, u=.5)
                t[ballot_high_u(ranking, self.voting_rule)] += self.have_ranking_with_utility_above_u(ranking, u=.5)
            else:
                t[ballot] += self.d_ranking_share[ranking]
        return TauVector(t, voting_rule=self.voting_rule)

    def is_equilibrium(self, strategy):
        """Whether a strategy is an equilibrium.

        Parameters
        ----------
        strategy : StrategyTwelve
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        EquilibriumStatus
            Whether `strategy` is an equilibrium in this profile.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> strategy = StrategyTwelve({'abc': 'ab', 'bac': 'b', 'cab': 'utility-dependent'})
            >>> profile.is_equilibrium(strategy)
            EquilibriumStatus.EQUILIBRIUM
        """
        d_ranking_best_response = self.tau(strategy).d_ranking_best_response
        status = EquilibriumStatus.EQUILIBRIUM
        for ranking, share in self.d_ranking_share.items():
            if share == 0:
                continue
            best_response = d_ranking_best_response[ranking]
            if best_response.ballot == INCONCLUSIVE:  # pragma: no cover
                status = min(status, EquilibriumStatus.INCONCLUSIVE)
            else:
                type_1 = ranking[:1] + '_' + ranking[1:]  # E.g. a_bc
                type_12 = ranking[:2] + '_' + ranking[2:]  # E.g. ab_c
                if best_response.ballot == UTILITY_DEPENDENT:
                    if self.voting_rule == APPROVAL:
                        best_ballot_1, best_ballot_12 = ballot_one(ranking), ballot_one_two(ranking)
                    elif self.voting_rule == PLURALITY:
                        best_ballot_1, best_ballot_12 = ballot_one(ranking), ballot_two(ranking)
                    elif self.voting_rule == ANTI_PLURALITY:
                        best_ballot_1, best_ballot_12 = ballot_one_three(ranking), ballot_one_two(ranking)
                    else:
                        raise NotImplementedError
                else:
                    best_ballot_1 = best_response.ballot
                    best_ballot_12 = best_response.ballot
                if self.d_type_share[type_1] > 0 and getattr(strategy, type_1) != best_ballot_1:
                    return EquilibriumStatus.NOT_EQUILIBRIUM
                if self.d_type_share[type_12] > 0 and getattr(strategy, type_12) != best_ballot_12:
                    return EquilibriumStatus.NOT_EQUILIBRIUM
        return status

    @cached_property
    def analyzed_strategies(self):
        """AnalyzedStrategies : Analyzed strategies of the profile.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> profile.analyzed_strategies
            Equilibria:
            <abc: a, bac: b, cab: ac> ==> b (FF)
            <abc: a, bac: ab, cab: c> ==> a (D)
            <abc: ab, bac: b, cab: utility-dependent> ==> b (FF)
            <BLANKLINE>
            Non-equilibria:
            <abc: a, bac: b, cab: c> ==> b (FF)
            <abc: a, bac: b, cab: utility-dependent> ==> b (FF)
            <abc: a, bac: ab, cab: ac> ==> a (D)
            <abc: a, bac: ab, cab: utility-dependent> ==> a (D)
            <abc: ab, bac: b, cab: c> ==> b (FF)
            <abc: ab, bac: b, cab: ac> ==> b (FF)
            <abc: ab, bac: ab, cab: c> ==> a, b (FF)
            <abc: ab, bac: ab, cab: ac> ==> a (D)
            <abc: ab, bac: ab, cab: utility-dependent> ==> a (D)
        """
        equilibria = []
        utility_dependent = []
        inconclusive = []
        non_equilibria = []

        def possible_strategies(share_ranking_1, share_ranking_12, strategy_1, strategy_12):
            if share_ranking_1 > 0 and share_ranking_12 > 0:
                return [strategy_1, strategy_12, UTILITY_DEPENDENT]
            elif share_ranking_1 > 0 or share_ranking_12 > 0:
                return [strategy_1, strategy_12]
            else:
                return ['']

        d_ranking_possible_strategies = {
            ranking: possible_strategies(share_ranking_1=self.d_type_share[ranking[0] + '_' + ranking[1:]],
                                         share_ranking_12=self.d_type_share[ranking[0:2] + '_' + ranking[2]],
                                         strategy_1=ballot_low_u(ranking, self.voting_rule),
                                         strategy_12=ballot_high_u(ranking, self.voting_rule))
            for ranking in RANKINGS
        }

        for d_ranking_strategy in product_dict(d_ranking_possible_strategies):
            strategy = StrategyTwelve(d_ranking_strategy, profile=self)
            status = strategy.is_equilibrium
            if status == EquilibriumStatus.EQUILIBRIUM:
                equilibria.append(strategy)
            elif status == EquilibriumStatus.UTILITY_DEPENDENT:  # pragma: no cover
                utility_dependent.append(strategy)
                warnings.warn('Met a utility-dependent case: \nprofile = %r\nstrategy = %r' % (self, strategy))
            elif status == EquilibriumStatus.INCONCLUSIVE:  # pragma: no cover
                inconclusive.append(strategy)
                warnings.warn('Met an inconclusive case: \nprofile = %r\nstrategy = %r' % (self, strategy))
            else:
                non_equilibria.append(strategy)
        return AnalyzedStrategies(equilibria, utility_dependent, inconclusive, non_equilibria)


def make_property_type_share(t, doc):
    def _f(self):
        return self.d_type_share[t]
    _f.__doc__ = doc
    return property(_f)


for my_t in TWELVE_TYPES:
    setattr(ProfileTwelve, my_t, make_property_type_share(my_t, 'Number : Share of voters with this type.'))



from poisson_approval.utils.DictPrintingInOrder import DictPrintingInOrder
from poisson_approval.containers.Winners import Winners


class Scores(DictPrintingInOrder):
    """The scores of the candidates

    Examples
    --------
        >>> scores = Scores({'a': 0.7, 'b':0.7, 'c':0.3})
        >>> print(scores)
        {a: 0.7, b: 0.7, c: 0.3}
        >>> print(scores.winners)
        a, b
    """

    @property
    def winners(self):
        """Winners: The set of winners."""
        return Winners(k for k, v in self.items() if v == max(self.values()))



import numpy as np
import pytest
from poisson_approval import Asymptotic, isnan, isposinf


def test_limit():
    assert isnan(Asymptotic(mu=np.nan, nu=51, xi=69).limit)
    assert isnan(Asymptotic(mu=0, nu=np.nan, xi=69).limit)
    assert isposinf(Asymptotic(mu=0, nu=51, xi=69).limit)


def test_add():
    asymptotic = Asymptotic(mu=np.nan, nu=51, xi=69) + Asymptotic(mu=42, nu=51, xi=69)
    assert isnan(asymptotic.mu)
    assert isnan(asymptotic.nu)
    assert isnan(asymptotic.xi)

    asymptotic = Asymptotic(mu=42, nu=51, xi=69) + Asymptotic(mu=42, nu=np.nan, xi=69)
    assert asymptotic.mu == 42
    assert isnan(asymptotic.nu)
    assert isnan(asymptotic.xi)


def test_isclose():
    assert not Asymptotic(mu=42, nu=51, xi=69).isclose(144)
    with pytest.raises(ValueError):
        Asymptotic(mu=np.nan, nu=51, xi=69).isclose(Asymptotic(mu=42, nu=51, xi=69))


def test_str_small_magnitudes():
    asymptotic = Asymptotic(mu=-2E-10, nu=0, xi=0)
    assert str(asymptotic) == 'exp(- 2e-10 n + o(1))'


def test_limit_ratio_with_small_magnitudes():
    numerator = Asymptotic(mu=-2E-10, nu=0, xi=0)
    denominator = Asymptotic(mu=-1E-10, nu=0, xi=0)
    ratio = numerator / denominator
    assert ratio.limit == 0



from fractions import Fraction
from poisson_approval import TauVector, BestResponseApproval, PLURALITY, ANTI_PLURALITY


def test_normalization():
    tau = TauVector({'a': 1, 'b': 1})
    assert tau.a == 0.5


def test_easy_vs_difficult_pivot():
    tau = TauVector({'a': Fraction(1, 3), 'ab': Fraction(1, 3), 'b': Fraction(1, 6), 'c': Fraction(1, 6)})
    assert tau.d_ranking_best_response['acb'].ballot == 'a'
    assert tau.d_ranking_best_response['acb'].justification == BestResponseApproval.EASY_VS_DIFFICULT
    assert tau.d_ranking_best_response['bca'].ballot == 'bc'
    assert tau.d_ranking_best_response['bca'].justification == BestResponseApproval.DIFFICULT_VS_EASY


def test_scores_in_duos():
    """
        >>> tau = TauVector({'a': Fraction(1, 10), 'ab': Fraction(3, 5), 'c': Fraction(3, 10)})
        >>> tau.score_ba_in_duo_ba
        0.6
        >>> tau.score_ca_in_duo_ca
        0.45825756949558394
        >>> tau.score_cb_in_duo_cb
        0.4242640687119285
        >>> tau.score_c_in_duo_ba
        Fraction(3, 10)
        >>> tau.score_b_in_duo_ca
        0.39279220242478624
        >>> tau.score_a_in_duo_cb
        0.5242640687119285
    """


def test_tau_plurality():
    """
        >>> tau = TauVector({'a': Fraction(2, 5), 'b': Fraction(2, 5), 'c': Fraction(1, 5)}, voting_rule=PLURALITY)
        >>> tau
        TauVector({'a': Fraction(2, 5), 'b': Fraction(2, 5), 'c': Fraction(1, 5)}, voting_rule='Plurality')
        >>> tau.d_ranking_best_response
        {'abc': <ballot = a, threshold_utility = 1, justification = Plurality analysis>, \
'acb': <ballot = a, threshold_utility = 1, justification = Plurality analysis>, \
'bac': <ballot = b, threshold_utility = 1, justification = Plurality analysis>, \
'bca': <ballot = b, threshold_utility = 1, justification = Plurality analysis>, \
'cab': <ballot = a, threshold_utility = 0, justification = Plurality analysis>, \
'cba': <ballot = b, threshold_utility = 0, justification = Plurality analysis>}
    """
    pass


def test_tau_anti_plurality():
    """
        >>> tau = TauVector({'ab': Fraction(2, 5), 'ac': Fraction(2, 5), 'bc': Fraction(1, 5)},
        ...                 voting_rule=ANTI_PLURALITY)
        >>> tau
        TauVector({'ab': Fraction(2, 5), 'ac': Fraction(2, 5), 'bc': Fraction(1, 5)}, voting_rule='Anti-plurality')
        >>> tau.d_ranking_best_response
        {'abc': <ballot = ab, threshold_utility = 0, justification = Anti-plurality analysis>, \
'acb': <ballot = ac, threshold_utility = 0, justification = Anti-plurality analysis>, \
'bac': <ballot = utility-dependent, threshold_utility = 0.5, justification = Anti-plurality analysis>, \
'bca': <ballot = bc, threshold_utility = 0, justification = Anti-plurality analysis>, \
'cab': <ballot = utility-dependent, threshold_utility = 0.5, justification = Anti-plurality analysis>, \
'cba': <ballot = bc, threshold_utility = 0, justification = Anti-plurality analysis>}
    """
    pass



import numpy as np
from math import isclose
from fractions import Fraction
from poisson_approval.constants.constants import *
from poisson_approval.utils.Util import isnan, ballot_high_u, ballot_low_u
from poisson_approval.utils.UtilCache import cached_property


class BestResponse:
    """Best response for a given ordinal type of voter (abstract class).

    The main objective of this class is to compute :attr:`threshold_utility`. The subclasses implement the
    best response in a specific voting rule.

    Parameters
    ----------
    tau : TauVector
        A tau-vector.
    ranking : str
        Voter's ranking, e.g. ``'abc'``.

    Attributes
    ----------
    i, j, k : str
        The first (resp. second, third) candidate in `ranking`. E.g. ``a``.
    ij, ik, jk : str
        The ballots with two candidates. E.g. ``ab``.
    tau_i, tau_j, tau_k, tau_ij, tau_ik, tau_jk : Number
        The values of the tau-vector.
    """

    def __init__(self, tau, ranking):
        self.tau = tau
        self.ranking = ranking
        self.i = ranking[0]
        self.j = ranking[1]
        self.k = ranking[2]
        self.ij = self.i + self.j
        self.ik = self.i + self.k
        self.ji = self.j + self.i
        self.jk = self.j + self.k
        self.ki = self.k + self.i
        self.kj = self.k + self.j
        self.tau_i = getattr(tau, self.i)
        self.tau_j = getattr(tau, self.j)
        self.tau_k = getattr(tau, self.k)
        self.tau_ij = getattr(tau, self.ij)
        self.tau_ik = getattr(tau, self.ik)
        self.tau_ji = getattr(tau, self.ji)
        self.tau_jk = getattr(tau, self.jk)
        self.tau_ki = getattr(tau, self.ki)
        self.tau_kj = getattr(tau, self.kj)

    voting_rule = None

    # ===============================
    # Shortcuts for the events of tau
    # ===============================

    # Duos
    # ----

    @cached_property
    def duo_ij(self):
        """EventDuo : The duo ij."""
        return getattr(self.tau, 'duo_' + self.ij)

    @cached_property
    def duo_ik(self):
        """EventDuo : The duo ik."""
        return getattr(self.tau, 'duo_' + self.ik)

    @cached_property
    def duo_ji(self):
        """EventDuo : The duo ji."""
        return getattr(self.tau, 'duo_' + self.ji)

    @cached_property
    def duo_jk(self):
        """EventDuo : The duo jk."""
        return getattr(self.tau, 'duo_' + self.jk)

    @cached_property
    def duo_ki(self):
        """EventDuo : The duo ki."""
        return getattr(self.tau, 'duo_' + self.ki)

    @cached_property
    def duo_kj(self):
        """EventDuo : The duo kj."""
        return getattr(self.tau, 'duo_' + self.kj)

    # Weak pivots
    # -----------

    @cached_property
    def pivot_weak_ij(self):
        """EventPivotWeak : The weak pivot ij."""
        return getattr(self.tau, 'pivot_weak_' + self.ij)

    @cached_property
    def pivot_weak_ik(self):
        """EventPivotWeak : The weak pivot ik."""
        return getattr(self.tau, 'pivot_weak_' + self.ik)

    @cached_property
    def pivot_weak_ji(self):
        """EventPivotWeak : The weak pivot ji."""
        return getattr(self.tau, 'pivot_weak_' + self.ji)

    @cached_property
    def pivot_weak_jk(self):
        """EventPivotWeak : The weak pivot jk."""
        return getattr(self.tau, 'pivot_weak_' + self.jk)

    @cached_property
    def pivot_weak_ki(self):
        """EventPivotWeak : The weak pivot ki."""
        return getattr(self.tau, 'pivot_weak_' + self.ki)

    @cached_property
    def pivot_weak_kj(self):
        """EventPivotWeak : The weak pivot kj."""
        return getattr(self.tau, 'pivot_weak_' + self.kj)

    # Strict pivots
    # -------------

    @cached_property
    def pivot_strict_ij(self):
        """EventPivotStrict: The strict pivot ij."""
        return getattr(self.tau, 'pivot_strict_' + self.ij)

    @cached_property
    def pivot_strict_ik(self):
        """EventPivotStrict: The strict pivot ik."""
        return getattr(self.tau, 'pivot_strict_' + self.ik)

    @cached_property
    def pivot_strict_ji(self):
        """EventPivotStrict: The strict pivot ji."""
        return getattr(self.tau, 'pivot_strict_' + self.ji)

    @cached_property
    def pivot_strict_jk(self):
        """EventPivotStrict: The strict pivot jk."""
        return getattr(self.tau, 'pivot_strict_' + self.jk)

    @cached_property
    def pivot_strict_ki(self):
        """EventPivotStrict: The strict pivot ki."""
        return getattr(self.tau, 'pivot_strict_' + self.ki)

    @cached_property
    def pivot_strict_kj(self):
        """EventPivotStrict: The strict pivot kj."""
        return getattr(self.tau, 'pivot_strict_' + self.kj)

    # Personalized pivots tij
    # -----------------------

    @cached_property
    def pivot_tij_ijk(self):
        """EventPivotTij: The first personalized pivot for voters ijk."""
        return getattr(self.tau, 'pivot_tij_' + self.i + self.j + self.k)

    @cached_property
    def pivot_tij_ikj(self):
        """EventPivotTij: The first personalized pivot for voters ikj."""
        return getattr(self.tau, 'pivot_tij_' + self.i + self.k + self.j)

    @cached_property
    def pivot_tij_jik(self):
        """EventPivotTij: The first personalized pivot for voters jik."""
        return getattr(self.tau, 'pivot_tij_' + self.j + self.i + self.k)

    @cached_property
    def pivot_tij_jki(self):
        """EventPivotTij: The first personalized pivot for voters jki."""
        return getattr(self.tau, 'pivot_tij_' + self.j + self.k + self.i)

    @cached_property
    def pivot_tij_kij(self):
        """EventPivotTij: The first personalized pivot for voters kij."""
        return getattr(self.tau, 'pivot_tij_' + self.k + self.i + self.j)

    @cached_property
    def pivot_tij_kji(self):
        """EventPivotTij: The first personalized pivot for voters kji."""
        return getattr(self.tau, 'pivot_tij_' + self.k + self.j + self.i)

    # Personalized pivots tjk
    # -----------------------

    @cached_property
    def pivot_tjk_ijk(self):
        """EventPivotTjk: The second personalized pivot for voters ijk."""
        return getattr(self.tau, 'pivot_tjk_' + self.i + self.j + self.k)

    @cached_property
    def pivot_tjk_ikj(self):
        """EventPivotTjk: The second personalized pivot for voters ikj."""
        return getattr(self.tau, 'pivot_tjk_' + self.i + self.k + self.j)

    @cached_property
    def pivot_tjk_jik(self):
        """EventPivotTjk: The second personalized pivot for voters jik."""
        return getattr(self.tau, 'pivot_tjk_' + self.j + self.i + self.k)

    @cached_property
    def pivot_tjk_jki(self):
        """EventPivotTjk: The second personalized pivot for voters jki."""
        return getattr(self.tau, 'pivot_tjk_' + self.j + self.k + self.i)

    @cached_property
    def pivot_tjk_kij(self):
        """EventPivotTjk: The second personalized pivot for voters kij."""
        return getattr(self.tau, 'pivot_tjk_' + self.k + self.i + self.j)

    @cached_property
    def pivot_tjk_kji(self):
        """EventPivotTjk: The second personalized pivot for voters kji."""
        return getattr(self.tau, 'pivot_tjk_' + self.k + self.j + self.i)

    # Shortcuts for the personalized pivots for voters ijk
    # ----------------------------------------------------

    @cached_property
    def pivot_tij(self):
        """EventPivotTij : The `personalized pivot` between candidates i and j. This is just another notation for
        :attr:`pivot_tij_ijk`.
        """
        return getattr(self.tau, 'pivot_tij_' + self.ranking)

    @cached_property
    def pivot_tjk(self):
        """EventPivotTjk : The `personalized pivot` between candidates j and k. This is just another notation for
        :attr:`pivot_tjk_ijk`.
        """
        return getattr(self.tau, 'pivot_tjk_' + self.ranking)

    # Trio
    # ----

    @cached_property
    def trio(self):
        """EventTrio : The 3-candidate tie."""
        return getattr(self.tau, 'trio')

    # Trio1t
    # ------

    @cached_property
    def trio_1t_i(self):
        """EventTrio1t : The first `personalized trio` (where candidate `i` has one vote less)."""
        return getattr(self.tau, 'trio_1t_' + self.i)

    @cached_property
    def trio_1t_j(self):
        """EventTrio1t : The first `personalized trio` (where candidate `j` has one vote less)."""
        return getattr(self.tau, 'trio_1t_' + self.j)

    @cached_property
    def trio_1t_k(self):
        """EventTrio1t : The first `personalized trio` (where candidate `k` has one vote less)."""
        return getattr(self.tau, 'trio_1t_' + self.k)

    # Trio2t
    # ------

    @cached_property
    def trio_2t_ij(self):
        """EventTrio2t: The second `personalized trio` (where candidates i and j have one vote less)."""
        return getattr(self.tau, 'trio_2t_' + self.ij)

    @cached_property
    def trio_2t_ik(self):
        """EventTrio2t: The second `personalized trio` (where candidates i and k have one vote less)."""
        return getattr(self.tau, 'trio_2t_' + self.ik)

    @cached_property
    def trio_2t_ji(self):
        """EventTrio2t: The second `personalized trio` (where candidates j and i have one vote less)."""
        return getattr(self.tau, 'trio_2t_' + self.ji)

    @cached_property
    def trio_2t_jk(self):
        """EventTrio2t: The second `personalized trio` (where candidates j and k have one vote less)."""
        return getattr(self.tau, 'trio_2t_' + self.jk)

    @cached_property
    def trio_2t_ki(self):
        """EventTrio2t: The second `personalized trio` (where candidates k and i have one vote less)."""
        return getattr(self.tau, 'trio_2t_' + self.ki)

    @cached_property
    def trio_2t_kj(self):
        """EventTrio2t: The second `personalized trio` (where candidates k and j have one vote less)."""
        return getattr(self.tau, 'trio_2t_' + self.kj)

    # Shortcuts for the personalized trios for voters ijk
    # ---------------------------------------------------

    @cached_property
    def trio_1t(self):
        """EventTrio1t : The first `personalized trio`. This is just another notation for :attr:`trio_1t_i`."""
        return getattr(self.tau, 'trio_1t_' + self.i)

    @cached_property
    def trio_2t(self):
        """EventTrio1t : The second `personalized trio`. This is just another notation for :attr:`trio_2t_ij`."""
        return getattr(self.tau, 'trio_2t_' + self.ij)

    # Easy and difficult pivots
    # -------------------------

    @cached_property
    def pivot_ij_easy_or_tight(self):
        """bool : True if the pivot `ij` is easy or tight, False if it is difficult."""
        return getattr(self.tau, 'pivot_%s_easy_or_tight' % (self.i + self.j))

    @cached_property
    def pivot_ik_easy_or_tight(self):
        """bool : True if the pivot `ik` is easy or tight, False if it is difficult."""
        return getattr(self.tau, 'pivot_%s_easy_or_tight' % (self.i + self.k))

    @cached_property
    def pivot_ji_easy_or_tight(self):
        """bool : True if the pivot `ji` is easy or tight, False if it is difficult."""
        return getattr(self.tau, 'pivot_%s_easy_or_tight' % (self.j + self.i))

    @cached_property
    def pivot_jk_easy_or_tight(self):
        """bool : True if the pivot `jk` is easy or tight, False if it is difficult."""
        return getattr(self.tau, 'pivot_%s_easy_or_tight' % (self.j + self.k))

    @cached_property
    def pivot_ki_easy_or_tight(self):
        """bool : True if the pivot `ki` is easy or tight, False if it is difficult."""
        return getattr(self.tau, 'pivot_%s_easy_or_tight' % (self.k + self.i))

    @cached_property
    def pivot_kj_easy_or_tight(self):
        """bool : True if the pivot `kj` is easy or tight, False if it is difficult."""
        return getattr(self.tau, 'pivot_%s_easy_or_tight' % (self.k + self.j))

    # =======
    # Results
    # =======

    @cached_property
    def results(self):
        """tuple (threshold_utility, justification) : Cf. :attr:`threshold_utility` and :attr:`justification`."""
        raise NotImplementedError

    @cached_property
    def threshold_utility(self):
        """Number : The threshold value of the utility for the second candidate (where the optimal ballot changes)."""
        return self.results[0]

    @cached_property
    def justification(self):
        """str : How the program computed the utility threshold."""
        return self.results[1]

    @cached_property
    def ballot(self):
        """str : This can be a valid ballot or ``'utility-dependent'``.
        """
        if isnan(self.threshold_utility):
            raise AssertionError('Unable to compute threshold utility')  # pragma: no cover
        elif isclose(self.threshold_utility, 1):
            return ballot_low_u(self.ranking, self.voting_rule)
        elif isclose(self.threshold_utility, 0, abs_tol=1E-9):
            return ballot_high_u(self.ranking, self.voting_rule)
        else:
            assert 0 <= self.threshold_utility <= 1
            return UTILITY_DEPENDENT

    def __repr__(self):
        return '<' + ', '.join([
            'ballot = %s' % self.ballot,
            'threshold_utility = {:.6g}'.format(float(self.threshold_utility)),
            'justification = %s' % self.justification,
        ]) + '>'

    @cached_property
    def _str_very_verbose(self):
        """Very verbose report.

        This is mostly used for testing purposes.

        Returns
        -------
        str
            A very verbose report.
        """
        s = ''
        s += 'tau = %s\n' % self.tau
        s += 'ranking = %s\n' % self.ranking
        s += 'voting_rule  = %s\n' % self.voting_rule
        s += 'duo_ij = %s\n' % self.duo_ij
        s += 'duo_ij = %s\n' % self.duo_ij
        s += 'duo_ji = %s\n' % self.duo_ji
        s += 'duo_ik = %s\n' % self.duo_ik
        s += 'duo_ki = %s\n' % self.duo_ki
        s += 'duo_jk = %s\n' % self.duo_jk
        s += 'duo_kj = %s\n' % self.duo_kj
        s += 'pivot_weak_ij = %s\n' % self.pivot_weak_ij
        s += 'pivot_weak_ji = %s\n' % self.pivot_weak_ji
        s += 'pivot_weak_ik = %s\n' % self.pivot_weak_ik
        s += 'pivot_weak_ki = %s\n' % self.pivot_weak_ki
        s += 'pivot_weak_jk = %s\n' % self.pivot_weak_jk
        s += 'pivot_weak_kj = %s\n' % self.pivot_weak_kj
        s += 'pivot_strict_ij = %s\n' % self.pivot_strict_ij
        s += 'pivot_strict_ji = %s\n' % self.pivot_strict_ji
        s += 'pivot_strict_ik = %s\n' % self.pivot_strict_ik
        s += 'pivot_strict_ki = %s\n' % self.pivot_strict_ki
        s += 'pivot_strict_jk = %s\n' % self.pivot_strict_jk
        s += 'pivot_strict_kj = %s\n' % self.pivot_strict_kj
        s += 'pivot_tij_ijk = %s\n' % self.pivot_tij_ijk
        s += 'pivot_tij_ikj = %s\n' % self.pivot_tij_ikj
        s += 'pivot_tij_jik = %s\n' % self.pivot_tij_jik
        s += 'pivot_tij_jki = %s\n' % self.pivot_tij_jki
        s += 'pivot_tij_kij = %s\n' % self.pivot_tij_kij
        s += 'pivot_tij_kji = %s\n' % self.pivot_tij_kji
        s += 'pivot_tjk_ijk = %s\n' % self.pivot_tjk_ijk
        s += 'pivot_tjk_ikj = %s\n' % self.pivot_tjk_ikj
        s += 'pivot_tjk_jik = %s\n' % self.pivot_tjk_jik
        s += 'pivot_tjk_jki = %s\n' % self.pivot_tjk_jki
        s += 'pivot_tjk_kij = %s\n' % self.pivot_tjk_kij
        s += 'pivot_tjk_kji = %s\n' % self.pivot_tjk_kji
        s += 'pivot_tij = %s\n' % self.pivot_tij
        s += 'pivot_tjk = %s\n' % self.pivot_tjk
        s += 'trio = %s\n' % self.trio
        s += 'trio_1t_i = %s\n' % self.trio_1t_i
        s += 'trio_1t_j = %s\n' % self.trio_1t_j
        s += 'trio_1t_k = %s\n' % self.trio_1t_k
        s += 'trio_2t_ij = %s\n' % self.trio_2t_ij
        s += 'trio_2t_ji = %s\n' % self.trio_2t_ji
        s += 'trio_2t_ik = %s\n' % self.trio_2t_ik
        s += 'trio_2t_ki = %s\n' % self.trio_2t_ki
        s += 'trio_2t_jk = %s\n' % self.trio_2t_jk
        s += 'trio_2t_kj = %s\n' % self.trio_2t_kj
        s += 'pivot_ij_easy_or_tight = %s\n' % self.pivot_ij_easy_or_tight
        s += 'pivot_ji_easy_or_tight = %s\n' % self.pivot_ji_easy_or_tight
        s += 'pivot_ik_easy_or_tight = %s\n' % self.pivot_ik_easy_or_tight
        s += 'pivot_ki_easy_or_tight = %s\n' % self.pivot_ki_easy_or_tight
        s += 'pivot_jk_easy_or_tight = %s\n' % self.pivot_jk_easy_or_tight
        s += 'pivot_kj_easy_or_tight = %s\n' % self.pivot_kj_easy_or_tight
        s += 'threshold_utility = %s\n' % self.threshold_utility
        s += 'justification = %s\n' % self.justification
        s += 'ballot = %s' % self.ballot
        return s



"""Top-level package for Poisson Approval."""

__author__ = """François Durand"""
__email__ = 'fradurand@gmail.com'
__version__ = '0.10.0'


# Utils
from poisson_approval.utils.DictPrintingInOrder import DictPrintingInOrder
from poisson_approval.utils.DictPrintingInOrderIgnoringNone import DictPrintingInOrderIgnoringNone
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros
from poisson_approval.utils.SetPrintingInOrder import SetPrintingInOrder
from poisson_approval.utils.Util import *
from poisson_approval.utils.UtilCache import *
from poisson_approval.utils.UtilMasks import *

# Constants
from poisson_approval.constants.constants import *
from poisson_approval.constants.EquilibriumStatus import EquilibriumStatus
from poisson_approval.constants.Focus import Focus

# Simple containers
from poisson_approval.containers.Winners import Winners
from poisson_approval.containers.Scores import Scores
from poisson_approval.containers.AnalyzedStrategies import AnalyzedStrategies

# Events
from poisson_approval.events.Asymptotic import Asymptotic
from poisson_approval.events.Event import Event
from poisson_approval.events.EventDuo import EventDuo
from poisson_approval.events.EventPivotStrict import EventPivotStrict
from poisson_approval.events.EventPivotTij import EventPivotTij
from poisson_approval.events.EventPivotTjk import EventPivotTjk
from poisson_approval.events.EventPivotWeak import EventPivotWeak
from poisson_approval.events.EventTrio import EventTrio
from poisson_approval.events.EventTrio1t import EventTrio1t
from poisson_approval.events.EventTrio2t import EventTrio2t

# Best response
from poisson_approval.best_response.BestResponse import BestResponse
from poisson_approval.best_response.BestResponseAntiPlurality import BestResponseAntiPlurality
from poisson_approval.best_response.BestResponseApproval import BestResponseApproval
from poisson_approval.best_response.BestResponsePlurality import BestResponsePlurality

# Tau-vector
from poisson_approval.tau_vector.TauVector import TauVector

# Strategies
from poisson_approval.strategies.Strategy import Strategy
from poisson_approval.strategies.StrategyTwelve import StrategyTwelve
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold
from poisson_approval.strategies.StrategyOrdinal import StrategyOrdinal

# Profiles
from poisson_approval.profiles.Profile import Profile
from poisson_approval.profiles.ProfileDiscrete import ProfileDiscrete
from poisson_approval.profiles.ProfileOrdinal import ProfileOrdinal
from poisson_approval.profiles.ProfileCardinal import ProfileCardinal
from poisson_approval.profiles.ProfileTwelve import ProfileTwelve
from poisson_approval.profiles.ProfileHistogram import ProfileHistogram

# Grid explorations
from poisson_approval.explore_grid.ExploreGridProfilesOrdinal import ExploreGridProfilesOrdinal
from poisson_approval.explore_grid.ExploreGridTaus import ExploreGridTaus

# Random generators
from poisson_approval.generators.GeneratorExample import GeneratorExample
from poisson_approval.generators.GeneratorProfileOrdinalGridUniform import GeneratorProfileOrdinalGridUniform
from poisson_approval.generators.GeneratorProfileOrdinalUniform import GeneratorProfileOrdinalUniform
from poisson_approval.generators.GeneratorProfileOrdinalVariations import GeneratorProfileOrdinalVariations
from poisson_approval.generators.GeneratorProfileHistogramUniform import GeneratorProfileHistogramUniform
from poisson_approval.generators.GeneratorStrategyOrdinalUniform import GeneratorStrategyOrdinalUniform
from poisson_approval.generators.GeneratorStrategyThresholdUniform import GeneratorStrategyThresholdUniform
from poisson_approval.generators.GeneratorTauVectorGridUniform import GeneratorTauVectorGridUniform
from poisson_approval.generators.GeneratorTauVectorUniform import GeneratorTauVectorUniform

# Meta-analysis
from poisson_approval.meta_analysis.NiceStatsProfileOrdinal import NiceStatsProfileOrdinal



from scipy.optimize import minimize
from math import sqrt
import numpy as np
from poisson_approval.events.Asymptotic import Asymptotic
from poisson_approval.events.Event import Event


class EventTrio(Event):
    """A 3-candidate tie.

    Notes
    -----
    We consider the 3-way tie, i.e. situations where ``S_x = S_y = S_z``.

    For parameters and attributes, cf. :class:`Event`.

    Examples
    --------
        >>> from fractions import Fraction
        >>> EventTrio(candidate_x='a', candidate_y='b', candidate_z='c',
        ...           tau_a=Fraction(1, 10), tau_ab=Fraction(6, 10), tau_c=Fraction(3, 10))
        <asymptotic = exp(- 0.151472 n - 0.5 log n - 0.836813 + o(1)), phi_a = 0, phi_c = 1.41421, phi_ab = 0.707107>
        >>> EventTrio(candidate_x='a', candidate_y='b', candidate_z='c',
        ...           tau_a=Fraction(1, 6), tau_b=Fraction(1, 6), tau_c=Fraction(1, 6),
        ...           tau_ab=Fraction(1, 6), tau_ac=Fraction(1, 6), tau_bc=Fraction(1, 6))
        <asymptotic = exp(? log n + ? + o(1)), phi_a = 1, phi_b = 1, phi_c = 1, phi_ab = 1, phi_ac = 1, phi_bc = 1>
    """

    def _compute(self, tau_x, tau_y, tau_z, tau_xy, tau_xz, tau_yz):
        is_cross_diagram = (tau_x == 0 and tau_yz == 0) or (tau_y == 0 and tau_xz == 0) or (tau_z == 0 and tau_xy == 0)
        is_flower_diagram = ((tau_x == 0 and tau_xy == 0) or (tau_x == 0 and tau_xz == 0)
                             or (tau_y == 0 and tau_xy == 0) or (tau_y == 0 and tau_yz == 0)
                             or (tau_z == 0 and tau_xz == 0) or (tau_z == 0 and tau_yz == 0))
        if is_cross_diagram or is_flower_diagram:
            self.asymptotic = (Asymptotic.poisson_eq(tau_x, tau_yz) * Asymptotic.poisson_eq(tau_y, tau_xz)
                               * Asymptotic.poisson_eq(tau_z, tau_xy))
            self._phi_x = sqrt(tau_yz / tau_x) if tau_x > 0 else np.nan
            self._phi_y = sqrt(tau_xz / tau_y) if tau_y > 0 else np.nan
            self._phi_z = sqrt(tau_xy / tau_z) if tau_z > 0 else np.nan
            self._phi_yz = sqrt(tau_x / tau_yz) if tau_yz > 0 else np.nan
            self._phi_xz = sqrt(tau_y / tau_xz) if tau_xz > 0 else np.nan
            self._phi_xy = sqrt(tau_z / tau_xy) if tau_xy > 0 else np.nan
        elif tau_xy == 0 and tau_xz == 0 and tau_yz == 0:
            # Tripod. Note that the other coefficient are not 0, otherwise it would be a (degenerate) cross diagram.
            self.asymptotic = Asymptotic(mu=3 * (tau_x * tau_y * tau_z) ** (1 / 3) - 1, nu=np.nan, xi=np.nan)
            self._phi_x = (tau_y * tau_z / tau_x ** 2) ** (1 / 3)
            self._phi_y = (tau_x * tau_z / tau_y ** 2) ** (1 / 3)
            self._phi_z = (tau_x * tau_y / tau_z ** 2) ** (1 / 3)
            self._phi_xy = np.nan
            self._phi_xz = np.nan
            self._phi_yz = np.nan
        elif tau_x == 0 and tau_y == 0 and tau_z == 0:
            # Inverted tripod. Note that the other coefficient are not 0, otherwise it would be a (degenerate) cross
            # diagram.
            self.asymptotic = Asymptotic(mu=3 * (tau_xy * tau_xz * tau_yz) ** (1 / 3) - 1, nu=np.nan, xi=np.nan)
            self._phi_x = np.nan
            self._phi_y = np.nan
            self._phi_z = np.nan
            self._phi_xy = (tau_xz * tau_yz / tau_xy ** 2) ** (1 / 3)
            self._phi_xz = (tau_xy * tau_yz / tau_xz ** 2) ** (1 / 3)
            self._phi_yz = (tau_xy * tau_xz / tau_yz ** 2) ** (1 / 3)
        elif tau_x - tau_yz == tau_y - tau_xz == tau_z - tau_xy:
            # Natural general tie.
            # This would work in the general case, but we can avoid numeric approximations easily!
            self.asymptotic = Asymptotic(mu=0, nu=np.nan, xi=np.nan)
            self._phi_x = 1 if tau_x > 0 else np.nan
            self._phi_y = 1 if tau_y > 0 else np.nan
            self._phi_z = 1 if tau_z > 0 else np.nan
            self._phi_xy = 1 if tau_xy > 0 else np.nan
            self._phi_xz = 1 if tau_xz > 0 else np.nan
            self._phi_yz = 1 if tau_yz > 0 else np.nan
        else:
            # Generic case
            optimizer = minimize(
                lambda x: 2 * sqrt(
                    (tau_x * x[0] + tau_xz) * (tau_yz / x[0] + tau_y)
                ) + tau_xy * x[0] + tau_z / x[0] - 1,
                np.array([1 + 1e-10, 1.]),  # We don't start exactly at 1. to avoid some (rare) bugs
                bounds=((1E-14, None), (1., 1.))
            )
            self.asymptotic = Asymptotic(mu=optimizer.fun, nu=np.nan, xi=np.nan)
            x_2 = optimizer.x[0]
            x_1 = sqrt((tau_yz / x_2 + tau_y) / (tau_x * x_2 + tau_xz))
            # print('x_2 =', x_2)
            # print('x_1 =', x_1)
            self._phi_x = x_1 * x_2 if tau_x > 0 else np.nan
            self._phi_y = 1 / x_1 if tau_y > 0 else np.nan
            self._phi_xz = x_1 if tau_xz > 0 else np.nan
            self._phi_yz = 1 / (x_1 * x_2) if tau_yz > 0 else np.nan
            self._phi_xy = x_2 if tau_xy > 0 else np.nan
            self._phi_z = 1 / x_2 if tau_z > 0 else np.nan



from poisson_approval import StrategyThreshold, PLURALITY, ANTI_PLURALITY, UTILITY_DEPENDENT


def test_strategy_threshold_plurality():
    """
        >>> strategy = StrategyThreshold({'abc': 0.7, 'bac': 1, 'cba': 0}, voting_rule=PLURALITY)
        >>> strategy
        StrategyThreshold({'abc': 0.7, 'bac': 1, 'cba': 0}, voting_rule='Plurality')
        >>> print(strategy)
        <abc: utility-dependent (0.7), bac: b, cba: b> (Plurality)
        >>> strategy.a_bc
        'a'
        >>> strategy.ab_c
        'b'
    """
    pass


def test_strategy_threshold_anti_plurality():
    """
        >>> strategy = StrategyThreshold({'abc': 0.7, 'bac': 1, 'cba': 0}, voting_rule=ANTI_PLURALITY)
        >>> strategy
        StrategyThreshold({'abc': 0.7, 'bac': 1, 'cba': 0}, voting_rule='Anti-plurality')
        >>> print(strategy)
        <abc: utility-dependent (0.7), bac: bc, cba: bc> (Anti-plurality)
        >>> strategy.a_bc
        'ac'
        >>> strategy.ab_c
        'ab'
    """
    pass



from math import sqrt
import numpy as np
from poisson_approval.events.Event import Event
from poisson_approval.events.Asymptotic import Asymptotic


class EventDuo(Event):
    """A 2-candidate tie.

    Notes
    -----
    We consider the situations where ``S_x = S_y``. It is not necessary a pivot, because ``z`` can
    have a greater score.

    For parameters and attributes, cf. :class:`Event`.

    Examples
    --------
        >>> from fractions import Fraction
        >>> EventDuo(candidate_x='a', candidate_y='b', candidate_z='c',
        ...          tau_a=Fraction(1, 10), tau_ab=Fraction(6, 10), tau_c=Fraction(3, 10))
        <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
    """

    def _compute(self, tau_x, tau_y, tau_z, tau_xy, tau_xz, tau_yz):
        w_x = tau_x + tau_xz
        w_y = tau_y + tau_yz
        self.asymptotic = Asymptotic.poisson_eq(w_x, w_y)
        self._phi_x = sqrt(w_y / w_x) if tau_x > 0 else np.nan
        self._phi_xz = sqrt(w_y / w_x) if tau_xz > 0 else np.nan
        self._phi_y = sqrt(w_x / w_y) if tau_y > 0 else np.nan
        self._phi_yz = sqrt(w_x / w_y) if tau_yz > 0 else np.nan
        self._phi_z = 1 if tau_z > 0 else np.nan
        self._phi_xy = 1 if tau_xy > 0 else np.nan



from math import isclose
from fractions import Fraction
from poisson_approval.constants.constants import *
from poisson_approval.utils.Util import ballot_one, ballot_two, ballot_one_two, ballot_one_three, barycenter, \
    to_callable, ballot_low_u, ballot_high_u
from poisson_approval.profiles.Profile import Profile
from poisson_approval.tau_vector.TauVector import TauVector
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros
from poisson_approval.constants.EquilibriumStatus import EquilibriumStatus
from poisson_approval.utils.UtilCache import cached_property, property_deleting_cache
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold


# noinspection PyUnresolvedReferences
class ProfileCardinal(Profile):
    """A cardinal profile of preference (abstract class).

    Parameters
    ----------
    ratio_sincere : Number
        The ratio of sincere voters, in the interval [0, 1]. This is used for :meth:`tau`.
    ratio_fanatic : Number
        The ratio of fanatic voters, in the interval [0, 1]. This is used for :meth:`tau`. The sum of `ratio_sincere`
        and `ratio_fanatic` must not exceed 1.
    voting_rule : str
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``.
    """

    def __init__(self, ratio_sincere=0, ratio_fanatic=0, voting_rule=APPROVAL):
        super().__init__(voting_rule=voting_rule)
        self.ratio_sincere = ratio_sincere
        self.ratio_fanatic = ratio_fanatic

    well_informed_voters = property_deleting_cache('_well_informed_voters')
    ratio_fanatic = property_deleting_cache('_ratio_fanatic')

    def have_ranking_with_utility_above_u(self, ranking, u):
        """Share of voters who have a given ranking and strictly above a given utility for their middle candidate.

        Parameters
        ----------
        ranking : str
            A ranking, e.g. ``'abc'``.
        u : Number
            A utility between 0 and 1 (included).

        Returns
        -------
        Number
            The share of voters who have ranking `ranking` and a utility for their middle candidate strictly greater
            than `u`.
        """
        raise NotImplementedError

    def have_ranking_with_utility_u(self, ranking, u):
        """Share of voters who have a given ranking and a given utility for their middle candidate.

        Parameters
        ----------
        ranking : str
            A ranking, e.g. ``'abc'``.
        u : Number
            A utility between 0 and 1 (included).

        Returns
        -------
        Number
            The share of voters who have ranking `ranking` and a utility for their middle candidate equal to `u`.
        """
        raise NotImplementedError

    def have_ranking_with_utility_below_u(self, ranking, u):
        """Share of voters who have a given ranking and strictly below a given utility for their middle candidate.

        Parameters
        ----------
        ranking : str
            A ranking, e.g. ``'abc'``.
        u : Number
            A utility between 0 and 1 (included).

        Returns
        -------
        Number
            The share of voters who have ranking `ranking` and a utility for their middle candidate strictly lower
            than `u`.
        """
        raise NotImplementedError

    @cached_property
    def d_ranking_share(self):
        return DictPrintingInOrderIgnoringZeros({
            ranking: self.have_ranking_with_utility_above_u(ranking, 0) + self.have_ranking_with_utility_u(ranking, 0)
            for ranking in RANKINGS
        })

    def __eq__(self, other):
        raise NotImplementedError

    @cached_property
    def standardized_version(self):
        raise NotImplementedError

    # Tau and strategy-related stuff

    def tau(self, strategy):
        """Tau-vector associated to a strategy, with partial sincere and fanatic voting.

        Parameters
        ----------
        strategy : an argument accepted by :meth:`tau_strategic`

        Returns
        -------
        TauVector
            A share :attr:`ratio_sincere` of the voters vote sincerely (in the sense of :attr:`tau_sincere`),
            a share :attr:`ratio_fanatic` vote only for their top candidate, and the rest of the voters
            vote strategically (in the sense of :meth:`tau_strategic`). In other words, this tau-vector
            is the barycenter of ``tau_sincere``, ``tau_fanatic`` and ``tau_strategic(strategy)``, with respective
            weights ``self.ratio_sincere``, ``self.ratio_fanatic`` and ``1 - self.ratio_sincere - self.ratio_fanatic``.
        """
        tau_sincere = self.tau_sincere
        tau_fanatic = self.tau_fanatic
        tau_strategic = self.tau_strategic(strategy)
        t = {ballot: barycenter(a=tau_strategic.d_ballot_share[ballot],
                                b=[tau_sincere.d_ballot_share[ballot], tau_fanatic.d_ballot_share[ballot]],
                                ratio_b=[self.ratio_sincere, self.ratio_fanatic])
             for ballot in BALLOTS_WITHOUT_INVERSIONS}
        return TauVector(t, voting_rule=self.voting_rule)

    @cached_property
    def tau_sincere(self):
        """Tau-vector associated to sincere voting.

        Returns
        -------
        TauVector
            * In Approval, all voters approve of their top candidate, and voters approve of their middle candidate if
              and only if their utility for her is strictly greater than 0.5.
            * In Plurality, all voters vote for their top candidate.
            * In Anti-plurality, all voters vote against their bottom candidate (i.e. for the other two).

        Notes
        -----
        In Plurality and Anti-plurality, sincere and fanatic voting are the same. They differ only in Approval.
        """
        t = {ballot: 0 for ballot in BALLOTS_WITHOUT_INVERSIONS}
        for ranking in RANKINGS:
            if self.voting_rule == APPROVAL:
                share_vote_one_two = self.have_ranking_with_utility_above_u(ranking, Fraction(1, 2))
                share_vote_one = self.d_ranking_share[ranking] - share_vote_one_two
                t[ballot_one(ranking)] += share_vote_one
                t[ballot_one_two(ranking)] += share_vote_one_two
            elif self.voting_rule == PLURALITY:
                t[ballot_one(ranking)] += self.d_ranking_share[ranking]
            elif self.voting_rule == ANTI_PLURALITY:
                t[ballot_one_two(ranking)] += self.d_ranking_share[ranking]
            else:
                raise NotImplementedError
        return TauVector(t, voting_rule=self.voting_rule)

    @cached_property
    def tau_fanatic(self):
        """Tau-vector associated to fanatic voting.

        Returns
        -------
        TauVector
            * In Approval or Plurality, all voters approve of their top candidate only.,
            * In Anti-plurality, all voters vote against their bottom candidate (i.e. for the other two).

        Notes
        -----
        In Plurality and Anti-plurality, sincere and fanatic voting are the same. They differ only in Approval.
        """
        t = {ballot: 0 for ballot in BALLOTS_WITHOUT_INVERSIONS}
        for ranking, share in self.d_ranking_share.items():
            if self.voting_rule in {APPROVAL, PLURALITY}:
                t[ballot_one(ranking)] += share
            elif self.voting_rule == ANTI_PLURALITY:
                t[ballot_one_two(ranking)] += self.d_ranking_share[ranking]
            else:
                raise NotImplementedError
        return TauVector(t, voting_rule=self.voting_rule)

    def tau_strategic(self, strategy):
        """Tau-vector associated to a strategy (fully strategic voting).

        Parameters
        ----------
        strategy : StrategyThreshold
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        TauVector
            Tau-vector associated to this profile and strategy `strategy`.
        """
        assert self.voting_rule == strategy.voting_rule
        t = {ballot: 0 for ballot in BALLOTS_WITHOUT_INVERSIONS}
        for ranking, threshold in strategy.d_ranking_threshold.items():
            if self.d_ranking_share[ranking] == 0:
                continue
            t[ballot_low_u(ranking, self.voting_rule)] += (
                self.have_ranking_with_utility_u(ranking, u=threshold)
                + self.have_ranking_with_utility_below_u(ranking, u=threshold))
            t[ballot_high_u(ranking, self.voting_rule)] += self.have_ranking_with_utility_above_u(ranking, u=threshold)
        return TauVector(t, voting_rule=self.voting_rule)

    def is_equilibrium(self, strategy):
        """Whether a strategy is an equilibrium.

        Parameters
        ----------
        strategy : StrategyThreshold
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        EquilibriumStatus
            Whether `strategy` is an equilibrium in this profile. This is based on the assumption that:

            * A proportion :attr:`ratio_sincere` of voters cast their ballot sincerely (in the sense of
              :attr:`tau_sincere`),
            * A proportion :attr:`ratio_fanatic` of voters vote for their top candidate only,
            * And the rest of the voters use `strategy`.
        """
        this_tau = self.tau(strategy)
        d_ranking_best_response = this_tau.d_ranking_best_response
        d_ranking_threshold = dict()
        for ranking, share in self.d_ranking_share.items():
            if share == 0:
                continue
            best_response = d_ranking_best_response[ranking]
            if best_response.ballot == INCONCLUSIVE:  # pragma: no cover
                return EquilibriumStatus.INCONCLUSIVE
            d_ranking_threshold[ranking] = best_response.threshold_utility
        tau_response = self.tau(StrategyThreshold(d_ranking_threshold))
        if tau_response.isclose(this_tau):
            return EquilibriumStatus.EQUILIBRIUM
        else:
            return EquilibriumStatus.NOT_EQUILIBRIUM

    def iterated_voting(self, strategy_ini, n_max_episodes,
                        perception_update_ratio=1, ballot_update_ratio=1, verbose=False):
        """Seek for convergence by iterated voting.

        Parameters
        ----------
        strategy_ini : an argument accepted by :meth:`tau`, i.e. by :meth:`tau_strategic`
            The initial strategy.
        n_max_episodes : int
            Maximal number of iterations.
        perception_update_ratio : Number in [0, 1]
            The coefficient when updating the perceived tau:
            ``tau_perceived = (1 - perception_update_ratio) * tau_perceived + perception_update_ratio * tau_actual``.
        ballot_update_ratio : Number in [0, 1]
            The ratio of voters who update their ballot:
            ``tau_actual = (1 - ballot_update_ratio) * tau_actual + ballot_update_ratio * tau_response``.
        verbose : bool
            If True, print all intermediate steps.

        Returns
        -------
        dict
            * Key ``cycle_taus_perceived``: list of :class:`TauVector`. The limit cycle of perceived tau-vectors.
              ``cycle_taus_perceived[t]`` is a barycenter of ``cycle_taus_perceived[t - 1]`` with
              ``cycle_taus_actual[t - 1]``, parametrized by `perception_update_ratio`.
            * Key ``cycle_strategies``: list of :class:`StrategyThreshold`. The limit cycle of strategies.
              ``cycle_strategies[t]`` is the best response to ``cycle_taus_perceived[t]``.
            * Key ``cycle_taus_actual``: list of :class:`TauVector`. The limit cycle of actual tau-vectors.
              ``cycle_taus_actual[t]`` is a barycenter of ``cycle_taus_actual[t - 1]`` and the tau-vector resulting
              from ``strategies[t]``, parametrized by `ballot_update_ratio`.
            * Key ``n_episodes``: the number of episodes until convergence. If the process did not converge, by
              convention, this value is `n_max_episodes`.

            `cycle_taus_perceived`, `cycle_strategies` and `cycle_taus_actual` have the same length. If it is 1, the
            process converges to this limit. If it is greater than 1, the process reaches a periodical orbit. If it
            is 0, by convention, it means that the process does not converge and does not reach a periodical orbit.

        Notes
        -----
        Comparison between :meth:`iterated_voting` and :meth:`fictitious_play`:

        * :meth:`iterated_voting` can detect cycles (whereas :meth:`fictitious_play` only looks for a limit).
        * :meth:`fictitious_play` accepts update ratios that are functions of the time (whereas
          :meth:`iterated_voting` only accepts constants).
        * :meth:`fictitious_play` is faster and uses less memory, because it only looks for a limit and not for a cycle.

        In general, you should use :meth:`iterated_voting` only if you care about cycles, with the constraint
        that it implies having constant update ratios.
        """
        strategy = strategy_ini.deepcopy_with_attached_profile(self)
        tau_actual = strategy.tau
        tau_perceived = tau_actual
        strategies = [strategy]
        taus_actual = [tau_actual]
        taus_perceived = [tau_perceived]
        if verbose:
            print('t = %s' % 0)
            print('strategy: %s' % strategy)
            print('tau_actual: %s' % tau_actual)
        n_episodes = n_max_episodes
        for t in range(1, n_max_episodes + 1):
            tau_perceived = TauVector({
                ballot: _my_round(barycenter(a=tau_perceived.d_ballot_share[ballot],
                                             b=tau_actual.d_ballot_share[ballot],
                                             ratio_b=perception_update_ratio))
                for ballot in BALLOTS_WITHOUT_INVERSIONS
            }, voting_rule=self.voting_rule)
            strategy = self.best_responses_to_strategy(tau_perceived.d_ranking_best_response)
            tau_full_response = strategy.tau
            tau_actual = TauVector({
                ballot: _my_round(barycenter(a=tau_actual.d_ballot_share[ballot],
                                             b=tau_full_response.d_ballot_share[ballot],
                                             ratio_b=ballot_update_ratio))
                for ballot in BALLOTS_WITHOUT_INVERSIONS
            }, normalization_warning=False, voting_rule=self.voting_rule)
            if verbose:
                print('t = %s' % t)
                print('tau_perceived: %s' % tau_perceived)
                print('strategy: %s' % strategy)
                print('tau_full_response: %s' % tau_full_response)
                print('tau_actual: %s' % tau_actual)
            # If there is an exact cycle, it is useless to continue looping.
            break_ = (tau_actual, tau_perceived) in zip(taus_actual, taus_perceived)

            taus_actual.append(tau_actual)
            taus_perceived.append(tau_perceived)
            strategies.append(strategy)
            if break_:
                n_episodes = t
                break
        try:
            end = len(taus_actual) - 1
            begin = next(begin
                         for begin in range(end - 1, -1, -1)
                         if taus_actual[begin].isclose(taus_actual[end], abs_tol=1E-9)
                         and taus_perceived[begin].isclose(taus_perceived[end], abs_tol=1E-9))
            cycle_taus_actual = taus_actual[begin + 1:end + 1]
            cycle_taus_perceived = taus_perceived[begin + 1:end + 1]
            cycle_strategies = strategies[begin + 1:end + 1]
        except StopIteration:
            cycle_taus_actual = []
            cycle_taus_perceived = []
            cycle_strategies = []
        return {'cycle_taus_perceived': cycle_taus_perceived,
                'cycle_strategies': cycle_strategies,
                'cycle_taus_actual': cycle_taus_actual,
                'n_episodes': n_episodes}

    def fictitious_play(self, strategy_ini, n_max_episodes,
                        perception_update_ratio=None, ballot_update_ratio=1, verbose=False):
        """Seek for convergence by fictitious play.

        Parameters
        ----------
        strategy_ini : an argument accepted by :meth:`tau`, i.e. by :meth:`tau_strategic`
            The initial strategy.
        n_max_episodes : int
            Maximal number of iterations.
        perception_update_ratio : callable or Number
            The coefficient when updating the perceived tau:
            ``tau_perceived = (1 - perception_update_ratio(t)) * tau_perceived + perception_update_ratio(t) *
            tau_actual``. For any ``t`` from 1 to `n_max_episodes` included, the update ratio must be in [0, 1]. The
            default function is ``1 / (t + 1)``, which leads to an arithmetic average. If `perception_update_ratio` is
            a Number, it is considered as a constant function.
        ballot_update_ratio : callable or Number
            The ratio of voters who update their ballot:
            ``tau_actual = (1 - ballot_update_ratio(t)) * tau_actual + ballot_update_ratio(t) * tau_response``.
            For any ``t`` from 1 to `n_max_episodes` included, the update ratio must be in [0, 1]. The default function
            is the constant 1, which corresponds to a full update. If `ballot_update_ratio` is a Number, it is
            considered as a constant function.
        verbose : bool
            If True, print all intermediate steps.

        Returns
        -------
        dict
            * Key ``tau``: :class:`TauVector` or None. The limit tau-vector. If None, it means that the process did not
              converge.
            * Key ``strategy``: :class:`StrategyThreshold` or None. The limit strategy. If None, it means that the
              process did not converge.
            * Key ``n_episodes``: the number of episodes until convergence. If the process did not converge, by
              convention, this value is `n_max_episodes`.

        Notes
        -----
        Comparison between :meth:`iterated_voting` and :meth:`fictitious_play`:

        * :meth:`iterated_voting` can detect cycles (whereas :meth:`fictitious_play` only looks for a limit).
        * :meth:`fictitious_play` accepts update ratios that are functions of the time (whereas
          :meth:`iterated_voting` only accepts constants).
        * :meth:`fictitious_play` is faster and uses less memory, because it only looks for a limit and not for a cycle.

        In general, you should use :meth:`iterated_voting` only if you care about cycles, with the constraint
        that it implies having constant update ratios.
        """
        if perception_update_ratio is None:
            def perception_update_ratio(_t): return Fraction(1, _t + 1)
        perception_update_ratio = to_callable(perception_update_ratio)
        ballot_update_ratio = to_callable(ballot_update_ratio)

        strategy = strategy_ini.deepcopy_with_attached_profile(self)
        tau_actual = strategy.tau
        tau_perceived = tau_actual
        if verbose:
            print('t = %s' % 0)
            print('strategy: %s' % strategy)
            print('tau_actual: %s' % tau_actual)
            print('tau_perceived: %s' % tau_perceived)

        for t in range(1, n_max_episodes + 1):
            tau_perceived = TauVector({
                ballot: _my_round(barycenter(a=tau_perceived.d_ballot_share[ballot],
                                             b=tau_actual.d_ballot_share[ballot],
                                             ratio_b=perception_update_ratio(t)))
                for ballot in BALLOTS_WITHOUT_INVERSIONS
            }, voting_rule=self.voting_rule)
            strategy = self.best_responses_to_strategy(tau_perceived.d_ranking_best_response)
            tau_full_response = strategy.tau
            tau_actual = TauVector({
                ballot: _my_round(barycenter(a=tau_actual.d_ballot_share[ballot],
                                             b=tau_full_response.d_ballot_share[ballot],
                                             ratio_b=ballot_update_ratio(t)))
                for ballot in BALLOTS_WITHOUT_INVERSIONS
            }, normalization_warning=False, voting_rule=self.voting_rule)
            if verbose:
                print('t = %s' % t)
                print('tau_perceived: %s' % tau_perceived)
                print('strategy: %s' % strategy)
                print('tau_full_response: %s' % tau_full_response)
                print('tau_actual: %s' % tau_actual)
            if tau_full_response.isclose(tau_perceived, abs_tol=1E-9) and tau_actual.isclose(
                    tau_full_response, abs_tol=1E-9):
                return {'tau': tau_full_response, 'strategy': strategy, 'n_episodes': t}
        return {'tau': None, 'strategy': None, 'n_episodes': n_max_episodes}


def _my_round(x):
    """Round to 0 or 1 if the number is close enough.

    Parameters
    ----------
    x : Number

    Returns
    -------
    Number
        Generally, return `x`. If `x` is close enough to 0 or 1, then return 0 or 1 respectively.

    Examples
    --------

        >>> _my_round(1E-10)
        0
        >>> _my_round(0.999999999999999999)
        1
        >>> _my_round(0.123456789)
        0.123456789
    """
    if isclose(x, 1):
        return 1
    if isclose(x, 0, abs_tol=1E-9):
        return 0
    return x



from math import sqrt, isclose
import numpy as np
from poisson_approval.events.Asymptotic import Asymptotic
from poisson_approval.events.Event import Event
from poisson_approval.events.EventTrio import EventTrio


class EventPivotStrict(Event):
    """A 2-candidate strict pivot.

    Notes
    -----
    We consider the strict pivot between ``x`` and ``y``, i.e. situations where ``S_x = S_y > S_z``.

    For parameters and attributes, cf. :class:`Event`.

    Examples
    --------
        >>> from fractions import Fraction
        >>> EventPivotStrict(candidate_x='a', candidate_y='b', candidate_z='c',
        ...                  tau_a=Fraction(1, 10), tau_ab=Fraction(6, 10), tau_c=Fraction(3, 10))
        <asymptotic = exp(- 0.1 n + o(1)), phi_a = 0, phi_c = 1, phi_ab = 1>
        >>> EventPivotStrict(candidate_x='a', candidate_y='b', candidate_z='c',
        ...                  tau_ab=Fraction(1, 10), tau_c=Fraction(9, 10))
        <asymptotic = exp(- 0.4 n - 0.5 log n - 1.35667 + o(1)), phi_c = 0.333333, phi_ab = 3>
        >>> EventPivotStrict(candidate_x='a', candidate_y='b', candidate_z='c',
        ...                  tau_a=Fraction(1, 6), tau_b=Fraction(1, 6), tau_c=Fraction(1, 6),
        ...                  tau_ab=Fraction(1, 6), tau_ac=Fraction(1, 6), tau_bc=Fraction(1, 6))
        <asymptotic = exp(- 0.5 log n - 1.40935 + o(1)), phi_a = 1, phi_b = 1, phi_c = 1, phi_ab = 1, phi_ac = 1, \
phi_bc = 1>
        >>> EventPivotStrict(candidate_x='a', candidate_y='b', candidate_z='c',
        ...                  tau_b=Fraction(1, 2), tau_ac=Fraction(1, 2))
        <asymptotic = exp(- inf)>

        >>> EventPivotStrict(candidate_x='a', candidate_y='b', candidate_z='c',
        ...                  tau_a=Fraction(1, 9), tau_b=Fraction(1, 9),
        ...                  tau_ab=Fraction(1, 9), tau_ac=Fraction(1, 3), tau_bc=Fraction(1, 3))
        <asymptotic = exp(- 0.0181103 n + ? log n + ? + o(1)), phi_a = 1.17454, phi_b = 1.17454, phi_ab = 1.37954, \
phi_ac = 0.851397, phi_bc = 0.851397>
    """

    def _compute(self, tau_x, tau_y, tau_z, tau_xy, tau_xz, tau_yz):
        if (tau_x == 0 and tau_xz == 0) or (tau_y == 0 and tau_yz == 0):
            # Consecutive holes on one side
            # Piv = P(X_x = 0) P(X_y = 0) P(X_xz = 0) P(X_yz = 0) P(X_xy > X_z)
            self.asymptotic = (Asymptotic.poisson_eq(tau_x, 0) * Asymptotic.poisson_eq(tau_y, 0)
                               * Asymptotic.poisson_eq(tau_xz, 0) * Asymptotic.poisson_eq(tau_yz, 0)
                               * Asymptotic.poisson_gt(tau_xy, tau_z))
            self._phi_x = 0 if tau_x > 0 else np.nan
            self._phi_y = 0 if tau_y > 0 else np.nan
            self._phi_xz = 0 if tau_xz > 0 else np.nan
            self._phi_yz = 0 if tau_yz > 0 else np.nan
            if tau_xy >= tau_z:
                self._phi_xy = 1 if tau_xy > 0 else np.nan
                self._phi_z = 1 if tau_z > 0 else np.nan
            else:
                self._phi_xy = sqrt(tau_z / tau_xy) if tau_xy > 0 else np.nan
                self._phi_z = sqrt(tau_xy / tau_z) if tau_z > 0 else np.nan
        else:
            w_x = tau_x + tau_xz  # > 0
            w_y = tau_y + tau_yz  # > 0
            s_x = tau_xy + tau_x * sqrt(w_y / w_x)
            s_z = tau_z + tau_yz * sqrt(w_x / w_y)
            if isclose(s_x, s_z):
                if tau_z != 0 or tau_xy != 0 or (tau_x != 0 and tau_y != 0 and tau_xz != 0 and tau_yz != 0):
                    self.asymptotic = Asymptotic.poisson_eq(w_x, w_y) * (1 / 2)
                    self._phi_x = sqrt(w_y / w_x) if tau_x > 0 else np.nan
                    self._phi_xz = sqrt(w_y / w_x) if tau_xz > 0 else np.nan
                    self._phi_y = sqrt(w_x / w_y) if tau_y > 0 else np.nan
                    self._phi_yz = sqrt(w_x / w_y) if tau_yz > 0 else np.nan
                    self._phi_z = 1 if tau_z > 0 else np.nan
                    self._phi_xy = 1 if tau_xy > 0 else np.nan
                else:
                    self.asymptotic = Asymptotic(mu=-np.inf, nu=-np.inf, xi=-np.inf)
                    self._phi_x = np.nan
                    self._phi_y = np.nan
                    self._phi_z = np.nan
                    self._phi_xy = np.nan
                    self._phi_xz = np.nan
                    self._phi_yz = np.nan
            elif s_x > s_z:
                # "Easy" pivot
                # P(piv_ab) ~ P(S_a = S_b)
                self.asymptotic = Asymptotic.poisson_eq(w_x, w_y)
                self._phi_x = sqrt(w_y / w_x) if tau_x > 0 else np.nan
                self._phi_xz = sqrt(w_y / w_x) if tau_xz > 0 else np.nan
                self._phi_y = sqrt(w_x / w_y) if tau_y > 0 else np.nan
                self._phi_yz = sqrt(w_x / w_y) if tau_yz > 0 else np.nan
                self._phi_z = 1 if tau_z > 0 else np.nan
                self._phi_xy = 1 if tau_xy > 0 else np.nan
            else:
                # "Difficult" pivot
                # mu_ab = mu_abc
                pivot_trio = EventTrio('x', 'y', 'z', tau_x=tau_x, tau_y=tau_y, tau_z=tau_z,
                                       tau_xy=tau_xy, tau_xz=tau_xz, tau_yz=tau_yz)
                self._phi_x = pivot_trio.phi['x']
                self._phi_y = pivot_trio.phi['y']
                self._phi_z = pivot_trio.phi['z']
                self._phi_xy = pivot_trio.phi['xy']
                self._phi_xz = pivot_trio.phi['xz']
                self._phi_yz = pivot_trio.phi['yz']
                if tau_z > 0:
                    _phi_z_tilde = self._phi_z
                else:
                    _phi_z_tilde = self._phi_xz * self._phi_yz
                if _phi_z_tilde == 0:
                    # The general formula for the asymptotic would be valid, but I prefer to forbid multiplication by 0
                    # in ``Asymptotic``, so that I have to validate such cases explicitly.
                    self.asymptotic = Asymptotic(mu=-np.inf, nu=-np.inf, xi=-np.inf)
                    # Also, the offsets are different in that case: they are not defined.
                    self._phi_x = np.nan
                    self._phi_y = np.nan
                    self._phi_z = np.nan
                    self._phi_xy = np.nan
                    self._phi_xz = np.nan
                    self._phi_yz = np.nan
                else:
                    self.asymptotic = pivot_trio.asymptotic * (_phi_z_tilde / (1 - _phi_z_tilde))



from poisson_approval.constants.constants import *
from poisson_approval.utils.Util import initialize_random_seeds, rand_simplex_grid
from poisson_approval.profiles.ProfileOrdinal import ProfileOrdinal


class GeneratorProfileOrdinalGridUniform:
    """A generator of ordinal profiles (:class:`ProfileOrdinal`), uniform on a grid.

    Parameters
    ----------
    denominator : int
        The coefficients of the profile will be fractions with this denominator.
    kwargs : keyword arguments
        These additional arguments will be passed directly to :class:`ProfileOrdinal`.

    Notes
    -----
    The profile is drawn uniformly on the points of the simplex whose coordinates are fractions with the given
    denominator.

    Examples
    --------
        >>> initialize_random_seeds()
        >>> generator = GeneratorProfileOrdinalGridUniform(denominator=100, well_informed_voters=False)
        >>> profile = generator()
        >>> print(profile)
        <abc: 1/50, acb: 23/100, bac: 17/50, cab: 11/50, cba: 19/100> (badly informed voters)
        >>> profile.well_informed_voters
        False
    """

    def __init__(self, denominator, **kwargs):
        self.denominator = denominator
        self.kwargs = kwargs

    def __call__(self):
        """
        Returns
        -------
        ProfileOrdinal
            A profile.
        """
        x = rand_simplex_grid(d=6, denominator=self.denominator)
        return ProfileOrdinal({ranking: x[i] for i, ranking in enumerate(RANKINGS)},
                              **self.kwargs)



import warnings
from math import isclose
from matplotlib import pyplot as plt
import numpy as np
from poisson_approval.constants.constants import *
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold
from poisson_approval.profiles.ProfileCardinal import ProfileCardinal
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros
from poisson_approval.utils.UtilCache import cached_property


class ProfileHistogram(ProfileCardinal):
    """A profile of preference with histogram distributions of utility.

    Parameters
    ----------
    d_ranking_share : dict
        E.g. ``{'abc': 0.4, 'cab': 0.6}``. ``d_ranking_share['abc']`` is the probability that a voter prefers candidate
        ``a``, then candidate ``b``, then candidate ``c``.
    d_ranking_histogram : dict
        Each key is a ranking, e.g. ``'abc'``. Each value is a list that represents a piecewise constant probability
        density function (PDF) of having a utility `u` for the middle candidate, e.g. ``b``. By convention, the list
        sums to 1 (contrary to the usual convention where the integral of the function would sum to 1).

        For example, if the list is ``[0.4, 0.3, 0.2, 0.1]``, it means that a fraction 0.4 of voters ``'abc'`` have a
        utility for ``b`` that is in the first quarter, i.e. between 0 and 0.25. These voters are uniformly distributed
        in this segment.
    normalization_warning : bool
        Whether a warning should be issued if the input distribution is not normalized.
    ratio_sincere : Number
        The ratio of sincere voters, in the interval [0, 1]. This is used for :meth:`tau`.
    ratio_fanatic : Number
        The ratio of fanatic voters, in the interval [0, 1]. This is used for :meth:`tau`. The sum of `ratio_sincere`
        and `ratio_fanatic` must not exceed 1.
    voting_rule : str
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``.

    Notes
    -----
    If the input distribution is not normalized, the profile will be normalized anyway and a warning is
    issued (unless `normalization_warning` is False).

    Examples
    --------
        >>> from fractions import Fraction
        >>> profile = ProfileHistogram(
        ...     {'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)},
        ...     {'abc': [1], 'bac': [1, 0], 'cab': [Fraction(2, 3), 0, 0, 0, 0, 0, 0, 0, 0, Fraction(1, 3)]})
        >>> profile  # doctest: +NORMALIZE_WHITESPACE
        ProfileHistogram({'abc': Fraction(1, 10), 'bac': Fraction(3, 5), 'cab': Fraction(3, 10)}, {'abc': array([1]), \
'bac': array([1, 0]), 'cab': array([Fraction(2, 3), 0, 0, 0, 0, 0, 0, 0, 0, Fraction(1, 3)],
            dtype=object)})
        >>> print(profile)
        <abc: 1/10 [1], bac: 3/5 [1 0], cab: 3/10 [Fraction(2, 3) 0 0 0 0 0 0 0 0 Fraction(1, 3)]> (Condorcet winner: b)
        >>> profile.abc
        Fraction(1, 10)
        >>> profile.d_ranking_share['abc']  # Alternate syntax for profile.abc
        Fraction(1, 10)
        >>> profile.weighted_maj_graph
        array([[0, Fraction(-1, 5), Fraction(2, 5)],
               [Fraction(1, 5), 0, Fraction(2, 5)],
               [Fraction(-2, 5), Fraction(-2, 5), 0]], dtype=object)
        >>> profile.condorcet_winners
        Winners({'b'})
        >>> profile.is_profile_condorcet
        1.0
        >>> profile.has_majority_favorite  # Is one candidate 'top' in a majority of ballots?
        True
        >>> profile.has_majority_ranking  # Does one ranking represent a majority of ballots?
        True
        >>> profile.is_single_peaked  # Is the profile single-peaked?
        True
        >>> profile.support_in_rankings
        {'abc', 'bac', 'cab'}
        >>> profile.is_generic_in_rankings  # Are all rankings there?
        False
        >>> strategy = StrategyThreshold({'abc': 0, 'bac': 1, 'cab': Fraction(1, 2)}, profile=profile)
        >>> print(profile.tau_sincere)
        <a: 1/20, ab: 1/20, ac: 1/10, b: 3/5, c: 1/5> ==> b
        >>> print(profile.tau_fanatic)
        <a: 1/10, b: 3/5, c: 3/10> ==> b
        >>> print(profile.tau_strategic(strategy))
        <ab: 1/10, ac: 1/10, b: 3/5, c: 1/5> ==> b
        >>> print(profile.tau(strategy))
        <ab: 1/10, ac: 1/10, b: 3/5, c: 1/5> ==> b
        >>> profile.is_equilibrium(strategy)
        EquilibriumStatus.EQUILIBRIUM
        >>> strategy_ini = StrategyThreshold({'abc': .5, 'bac': .5, 'cab': .5})
        >>> cycle = profile.iterated_voting(strategy_ini, 100)['cycle_strategies']
        >>> len(cycle)
        1
        >>> print(cycle[0])
        <abc: ab, bac: utility-dependent (0.7199316142046179), cab: utility-dependent (0.28006838579538196)> ==> b
        >>> limit_strategy = profile.fictitious_play(strategy_ini, 100, perception_update_ratio=1)['strategy']
        >>> print(limit_strategy)
        <abc: ab, bac: utility-dependent (0.7199316142046179), cab: utility-dependent (0.28006838579538196)> ==> b
    """

    def __init__(self, d_ranking_share, d_ranking_histogram, normalization_warning=True,
                 ratio_sincere=0, ratio_fanatic=0, voting_rule=APPROVAL):
        """
            >>> profile = ProfileHistogram(d_ranking_share={'abc': 1},
            ...                            d_ranking_histogram={'non_existing_ranking': [1]})
            Traceback (most recent call last):
            KeyError: 'non_existing_ranking'
        """
        super().__init__(ratio_sincere=ratio_sincere, ratio_fanatic=ratio_fanatic, voting_rule=voting_rule)
        # Populate the dictionary (and check for typos in the input)
        self._d_ranking_share = DictPrintingInOrderIgnoringZeros({ranking: 0 for ranking in RANKINGS})
        self.d_ranking_histogram = DictPrintingInOrderIgnoringZeros({ranking: np.array([]) for ranking in RANKINGS})
        for ranking, share in d_ranking_share.items():
            self._d_ranking_share[ranking] += share
        for ranking, histogram in d_ranking_histogram.items():
            if ranking in RANKINGS:
                self.d_ranking_histogram[ranking] = np.array(histogram)
            else:
                raise KeyError('%s' % ranking)
        # Normalize if necessary
        total = sum(self._d_ranking_share.values())
        if not isclose(total, 1.):
            if normalization_warning:
                warnings.warn("Warning: profile is not normalized, I will normalize it.")
            for ranking in self._d_ranking_share.keys():
                self._d_ranking_share[ranking] /= total
        for ranking, histogram in self.d_ranking_histogram.items():
            if len(histogram) == 0:
                continue
            total = np.sum(histogram)
            if not isclose(total, 1.):
                if normalization_warning:
                    warnings.warn("Warning: profile is not normalized, I will normalize it.")
                self.d_ranking_histogram[ranking] = histogram / total

    @cached_property
    def d_ranking_share(self):
        return self._d_ranking_share

    def have_ranking_with_utility_above_u(self, ranking, u):
        """Share of voters who have a given ranking and strictly above a given utility for their middle candidate.

        Cf. :meth:`ProfileCardinal.have_ranking_with_utility_above_u`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileHistogram(
            ...     {'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)},
            ...     {'abc': [1], 'bac': [1, 0], 'cab': [Fraction(2, 3), 0, 0, 0, 0, 0, 0, 0, 0, Fraction(1, 3)]})
            >>> profile.have_ranking_with_utility_above_u(ranking='cab', u=0)
            Fraction(3, 10)
            >>> profile.have_ranking_with_utility_above_u(ranking='cab', u=Fraction(1, 100))
            Fraction(7, 25)
            >>> profile.have_ranking_with_utility_above_u(ranking='cab', u=Fraction(99, 100))
            Fraction(1, 100)
            >>> profile.have_ranking_with_utility_above_u(ranking='cab', u=1)
            Fraction(0, 1)
        """
        return self.d_ranking_share[ranking] - self.have_ranking_with_utility_below_u(ranking, u)

    def have_ranking_with_utility_u(self, ranking, u):
        """Share of voters who have a given ranking and a given utility for their middle candidate.

        Cf. :meth:`ProfileCardinal.have_ranking_with_utility_u`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileHistogram(
            ...     {'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)},
            ...     {'abc': [1], 'bac': [1, 0], 'cab': [Fraction(2, 3), 0, 0, 0, 0, 0, 0, 0, 0, Fraction(1, 3)]})
            >>> profile.have_ranking_with_utility_u(ranking='cab', u=Fraction(1, 100))
            0
        """
        return 0

    def have_ranking_with_utility_below_u(self, ranking, u):
        """Share of voters who have a given ranking and strictly below a given utility for their middle candidate.

        Cf. :meth:`ProfileCardinal.have_ranking_with_utility_below_u`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileHistogram(
            ...     {'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)},
            ...     {'abc': [1], 'bac': [1, 0], 'cab': [Fraction(2, 3), 0, 0, 0, 0, 0, 0, 0, 0, Fraction(1, 3)]})
            >>> profile.have_ranking_with_utility_below_u(ranking='cab', u=0)
            Fraction(0, 1)
            >>> profile.have_ranking_with_utility_below_u(ranking='cab', u=Fraction(1, 100))
            Fraction(1, 50)
            >>> profile.have_ranking_with_utility_below_u(ranking='cab', u=Fraction(99, 100))
            Fraction(29, 100)
            >>> profile.have_ranking_with_utility_below_u(ranking='cab', u=1)
            Fraction(3, 10)
        """
        share_ranking = self.d_ranking_share[ranking]
        if share_ranking == 0:
            return 0
        if u == 1:
            return share_ranking
        histogram = self.d_ranking_histogram[ranking]
        n_bins = len(histogram)
        k = int(u * n_bins)
        if histogram[k] == 0:
            # Not really an exception, but handles fractions more nicely.
            return share_ranking * np.sum(histogram[0:k])
        else:
            return share_ranking * (np.sum(histogram[0:k]) + histogram[k] * (u * n_bins - k))

    def __repr__(self):
        """
            >>> from fractions import Fraction
            >>> profile = ProfileHistogram(
            ...     {'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)},
            ...     {'abc': [1], 'bac': [1, 0], 'cab': [Fraction(2, 3), 0, 0, 0, 0, 0, 0, 0, 0, Fraction(1, 3)]},
            ...     ratio_sincere=Fraction(1, 10), ratio_fanatic=Fraction(2, 10))
            >>> profile
            ProfileHistogram({'abc': Fraction(1, 10), 'bac': Fraction(3, 5), 'cab': Fraction(3, 10)}, \
{'abc': array([1]), 'bac': array([1, 0]), 'cab': array([Fraction(2, 3), 0, 0, 0, 0, 0, 0, 0, 0, Fraction(1, 3)],
                  dtype=object)}, ratio_sincere=Fraction(1, 10), ratio_fanatic=Fraction(1, 5))
        """
        arguments = '%r, %r' % (self.d_ranking_share, self.d_ranking_histogram)
        if self.ratio_sincere > 0:
            arguments += ', ratio_sincere=%r' % self.ratio_sincere
        if self.ratio_fanatic > 0:
            arguments += ', ratio_fanatic=%r' % self.ratio_fanatic
        if self.voting_rule != APPROVAL:
            arguments += ', voting_rule=%r' % self.voting_rule
        return 'ProfileHistogram(%s)' % arguments

    def __str__(self):
        """
            >>> from fractions import Fraction
            >>> profile = ProfileHistogram(
            ...     {'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)},
            ...     {'abc': [1], 'bac': [1, 0], 'cab': [Fraction(2, 3), 0, 0, 0, 0, 0, 0, 0, 0, Fraction(1, 3)]},
            ...     ratio_sincere=Fraction(1, 10), ratio_fanatic=Fraction(2, 10))
            >>> print(profile)
            <abc: 1/10 [1], bac: 3/5 [1 0], cab: 3/10 [Fraction(2, 3) 0 0 0 0 0 0 0 0 Fraction(1, 3)]> \
(Condorcet winner: b) (ratio_sincere: 1/10) (ratio_fanatic: 1/5)
        """
        result = '<' + ', '.join([
            '%s: %s %s' % (ranking, self.d_ranking_share[ranking], self.d_ranking_histogram[ranking])
            for ranking in sorted(self.d_ranking_share)
            if self.d_ranking_share[ranking] > 0 or len(self.d_ranking_histogram[ranking]) > 0
        ]) + '>'
        if self.is_profile_condorcet:
            result += ' (Condorcet winner: %s)' % self.condorcet_winners
        if self.ratio_sincere > 0:
            result += ' (ratio_sincere: %s)' % self.ratio_sincere
        if self.ratio_fanatic > 0:
            result += ' (ratio_fanatic: %s)' % self.ratio_fanatic
        if self.voting_rule != APPROVAL:
            result += ' (%s)' % self.voting_rule
        return result

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')

    def __eq__(self, other):
        """Equality test.

        Parameters
        ----------
        other : Object

        Returns
        -------
        bool
            True iff this profile is equal to `other`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileHistogram(
            ...     {'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)},
            ...     {'abc': [1], 'bac': [1, 0], 'cab': [Fraction(2, 3), 0, 0, 0, 0, 0, 0, 0, 0, Fraction(1, 3)]})
            >>> profile == ProfileHistogram(
            ...     {'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)},
            ...     {'abc': [1], 'bac': [1, 0], 'cab': [Fraction(2, 3), 0, 0, 0, 0, 0, 0, 0, 0, Fraction(1, 3)]})
            True
        """
        return (isinstance(other, ProfileHistogram)
                and self.d_ranking_share == other.d_ranking_share
                and self.d_ranking_histogram == self.d_ranking_histogram
                and self.ratio_sincere == other.ratio_sincere
                and self.ratio_fanatic == other.ratio_fanatic
                and self.voting_rule == other.voting_rule)

    # Standardized version of the profile (makes it unique, up to permutations)

    @cached_property
    def standardized_version(self):
        """ProfileHistogram : Standardized version of the profile (makes it unique, up to permutations of the candidates).

         Examples
         --------
            >>> from fractions import Fraction
            >>> profile = ProfileHistogram(
            ...     {'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)},
            ...     {'abc': [1], 'bac': [1, 0], 'cab': [Fraction(2, 3), 0, 0, 0, 0, 0, 0, 0, 0, Fraction(1, 3)]})
            >>> print(profile.standardized_version)
            <abc: 3/5 [1 0], bac: 1/10 [1], cba: 3/10 [Fraction(2, 3) 0 0 0 0 0 0 0 0 Fraction(1, 3)]> \
(Condorcet winner: a)
            >>> profile.is_standardized
            False
        """
        def translate(s, permute):
            return s.replace('a', permute[0]).replace('b', permute[1]).replace('c', permute[2])

        best_d_ranking_share = {}
        best_d_ranking_histogram = {}
        best_signature = []
        for perm in XYZ_PERMUTATIONS:
            d_ranking_share_test = {translate(ranking, perm): share
                                    for ranking, share in self.d_ranking_share.items()}
            d_ranking_histogram_test = {translate(ranking, perm): histogram
                                        for ranking, histogram in self.d_ranking_histogram.items()}
            signature_test = [d_ranking_share_test[ranking] for ranking in XYZ_RANKINGS]
            for ranking in XYZ_RANKINGS:
                signature_test.extend(d_ranking_histogram_test[ranking])
            if signature_test > best_signature:
                best_signature = signature_test
                best_d_ranking_share = d_ranking_share_test
                best_d_ranking_histogram = d_ranking_histogram_test
        return ProfileHistogram(
            d_ranking_share={ranking: best_d_ranking_share[xyz_ranking]
                             for ranking, xyz_ranking in zip(RANKINGS, XYZ_RANKINGS)},
            d_ranking_histogram={ranking: best_d_ranking_histogram[xyz_ranking]
                                 for ranking, xyz_ranking in zip(RANKINGS, XYZ_RANKINGS)},
            ratio_sincere=self.ratio_sincere, ratio_fanatic=self.ratio_fanatic, voting_rule=self.voting_rule
        )

    def plot_cdf(self, ranking, x_label=None, y_label=None, **kwargs):
        """Plot the cumulative distribution function (CDF) for a given ranking.

        Parameters
        ----------
        ranking : str
            A ranking.
        x_label : str, optional
            The label for x-axis. If not specified, an appropriate label is provided.
        y_label
            The label for y-axis. If not specified, an appropriate label is provided.
        kwargs
            The additional keyword arguments are passed to :meth:`pyplot.plot`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileHistogram(
            ...     {'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)},
            ...     {'abc': [1], 'bac': [1, 0], 'cab': [Fraction(2, 3), 0, 0, 0, 0, 0, 0, 0, 0, Fraction(1, 3)]})
            >>> profile.plot_cdf('cab')
        """
        if x_label is None:
            x_label = 'Utility for %s' % ranking[1]
        if y_label is None:
            y_label = 'Cumulative proportion of the voters %s' % ranking
        n_bins = len(self.d_ranking_histogram[ranking])
        x = np.array(range(0, n_bins + 1)) / n_bins
        y = np.concatenate(([0], np.cumsum(self.d_ranking_histogram[ranking])))
        plt.plot(x, y, **kwargs)
        plt.xlabel(x_label)
        plt.ylabel(y_label)

    def plot_histogram(self, ranking, x_label=None, y_label=None, **kwargs):
        """Plot the histogram for a given ranking.

        Up to a renormalization, it is the probability density function (PDF).

        Parameters
        ----------
        ranking : str
            A ranking.
        x_label : str, optional
            The label for x-axis. If not specified, an appropriate label is provided.
        y_label
            The label for y-axis. If not specified, an appropriate label is provided.
        kwargs
            The additional keyword arguments are passed to :meth:`pyplot.plot`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileHistogram(
            ...     {'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)},
            ...     {'abc': [1], 'bac': [1, 0], 'cab': [Fraction(2, 3), 0, 0, 0, 0, 0, 0, 0, 0, Fraction(1, 3)]})
            >>> profile.plot_histogram('cab')
        """
        if x_label is None:
            x_label = 'Utility for %s' % ranking[1]
        if y_label is None:
            y_label = 'Proportion of the voters %s' % ranking
        n_bins = len(self.d_ranking_histogram[ranking])
        x = np.array(range(0, n_bins + 1)) / n_bins
        y = np.concatenate(([0], self.d_ranking_histogram[ranking]))
        plt.step(x, y, **kwargs)
        plt.xlabel(x_label)
        plt.ylabel(y_label)



from poisson_approval.constants.constants import *
from poisson_approval.utils.Util import initialize_random_seeds, rand_simplex
from poisson_approval.profiles.ProfileHistogram import ProfileHistogram


class GeneratorProfileHistogramUniform:
    """A generator of histogram-profiles (:class:`ProfileHistogram`) following the uniform distribution.

    Parameters
    ----------
    n_bins : int
        The numbers of bins for each histogram.
    kwargs : keyword arguments
        These additional arguments will be passed directly to :class:`ProfileHistogram`.

    Notes
    -----
    The ordinal profile is drawn uniformly on the simplex. Then each histogram is drawn uniformly on the simplex.

    Examples
    --------
        >>> initialize_random_seeds()
        >>> generator = GeneratorProfileHistogramUniform(n_bins=3)
        >>> profile = generator()
        >>> print(profile)
        <abc: 0.4236547993389047 [0.43758721 0.2083069  0.35410589], \
acb: 0.12122838365799216 [0.891773   0.07188976 0.03633724], \
bac: 0.0039303209304278885 [0.38344152 0.40828352 0.20827496], \
bca: 0.05394987214431912 [0.52889492 0.03914964 0.43195544], \
cab: 0.1124259903007756 [0.07103606 0.85456058 0.07440336], \
cba: 0.2848106336275805 [0.0202184 0.0669109 0.9128707]> (Condorcet winner: a)
    """

    def __init__(self, n_bins, **kwargs):
        self.n_bins = n_bins
        self.kwargs = kwargs

    def __call__(self):
        """
        Returns
        -------
        ProfileHistogram
            A profile.
        """
        x = rand_simplex(d=6)
        return ProfileHistogram({ranking: x[i] for i, ranking in enumerate(RANKINGS)},
                                {ranking: rand_simplex(d=self.n_bins) for ranking in RANKINGS},
                                **self.kwargs)



from poisson_approval.constants.constants import *
from poisson_approval.utils.Util import initialize_random_seeds
from poisson_approval.profiles.ProfileOrdinal import ProfileOrdinal
import random


class GeneratorProfileOrdinalVariations:
    """A generator of ordinal profiles (:class:`ProfileOrdinal`) describing variations of a given base profile.

    Parameters
    ----------
    base_profile : ProfileOrdinal
        The base profile.
    epsilon : Number
        The noise.
    kwargs : keyword arguments
        These additional arguments will be passed directly to :class:`ProfileOrdinal`.

    Notes
    -----
    We add ``epsilon * random()`` to each component of the base profile. Then we normalize to have a sum of 1.

    Examples
    --------
        >>> initialize_random_seeds()
        >>> base_profile = ProfileOrdinal({'abc': 0.25, 'bac': 0.75})
        >>> generator = GeneratorProfileOrdinalVariations(base_profile, epsilon=0.01)
        >>> profile = generator()
        >>> print(profile)  # doctest: +SKIP
        <abc: 0.25043511946419417, acb: 0.007344656519664636, bac: 0.7308331353826986, bca: 0.002508930076416972, \
cab: 0.004954304904190167, cba: 0.003923853652835374> (Condorcet winner: b)
    """

    def __init__(self, base_profile, epsilon, **kwargs):
        self.base_profile = base_profile
        self.epsilon = epsilon
        self.kwargs = kwargs

    def __call__(self):
        """
        :return: a profile
        """
        return ProfileOrdinal({ranking: self.base_profile.d_ranking_share[ranking] + self.epsilon * random.random()
                               for ranking in RANKINGS},
                              normalization_warning=False,
                              **self.kwargs)



from poisson_approval.constants.constants import *
from poisson_approval.utils.Util import initialize_random_seeds, rand_simplex
from poisson_approval.profiles.ProfileOrdinal import ProfileOrdinal


class GeneratorProfileOrdinalUniform:
    """A generator of ordinal profiles (:class:`ProfileOrdinal`) following the uniform distribution.

    Parameters
    ----------
    kwargs : keyword arguments
        These additional arguments will be passed directly to :class:`ProfileOrdinal`.

    Notes
    -----
    The profile is drawn uniformly on the simplex.

    Examples
    --------
        >>> initialize_random_seeds()
        >>> generator = GeneratorProfileOrdinalUniform(well_informed_voters=True)
        >>> profile = generator()
        >>> print(profile)
        <abc: 0.4236547993389047, acb: 0.12122838365799216, bac: 0.0039303209304278885, bca: 0.05394987214431912, \
cab: 0.1124259903007756, cba: 0.2848106336275805> (Condorcet winner: a)
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self):
        """
        Returns
        -------
        ProfileOrdinal
            A profile.
        """
        x = rand_simplex(d=6)
        return ProfileOrdinal({ranking: x[i] for i, ranking in enumerate(RANKINGS)},
                              **self.kwargs)



from poisson_approval import ProfileDiscrete, PLURALITY, ANTI_PLURALITY


def test_normalization():
    profile = ProfileDiscrete({('abc', 0.2): 1, ('acb', 0.4): 1})
    assert profile.abc == 0.5


def test_plurality():
    """
        >>> profile = ProfileDiscrete({('abc', 0.2): 1, ('acb', 0.4): 1}, voting_rule=PLURALITY)
        >>> profile
        ProfileDiscrete({'abc': {0.2: 0.5}, 'acb': {0.4: 0.5}}, voting_rule='Plurality')
        >>> print(profile)
        <abc 0.2: 0.5, acb 0.4: 0.5> (Condorcet winner: a) (Plurality)
    """
    pass


def test_anti_plurality():
    """
        >>> profile = ProfileDiscrete({('abc', 0.2): 1, ('acb', 0.4): 1}, voting_rule=ANTI_PLURALITY)
        >>> profile
        ProfileDiscrete({'abc': {0.2: 0.5}, 'acb': {0.4: 0.5}}, voting_rule='Anti-plurality')
        >>> print(profile)
        <abc 0.2: 0.5, acb 0.4: 0.5> (Condorcet winner: a) (Anti-plurality)
    """
    pass



import warnings
import numpy as np
from math import isclose
from poisson_approval.constants.constants import *
from poisson_approval.constants.EquilibriumStatus import EquilibriumStatus
from poisson_approval.containers.AnalyzedStrategies import AnalyzedStrategies
from poisson_approval.profiles.Profile import Profile
from poisson_approval.strategies.StrategyOrdinal import StrategyOrdinal
from poisson_approval.tau_vector.TauVector import TauVector
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros
from poisson_approval.utils.Util import ballot_one, ballot_two, ballot_one_two, ballot_one_three, barycenter, \
    product_dict, ballot_low_u, ballot_high_u
from poisson_approval.utils.UtilCache import cached_property, property_deleting_cache
from poisson_approval.utils.UtilMasks import masks_area, masks_distribution, winners_distribution


# noinspection PyUnresolvedReferences
class ProfileOrdinal(Profile):
    """An ordinal profile of preference.

    Parameters
    ----------
    d_ranking_share : dict
        E.g. ``{'abc': 0.4, 'cab': 0.6}``. ``d_ranking_share['abc']`` is the probability that a voter prefers
        candidate ``a``, then candidate ``b``, then candidate ``c``.
    normalization_warning : bool
        Whether a warning should be issued if the input distribution is not normalized.
    well_informed_voters : bool.
        If True (default), it is the usual model. If False, voters "see" only the candidates' expected scores and
        believe that the scores follow independent Poisson distributions. This option has an effect only for Approval
        (neither for Plurality nor Anti-plurality).
    ratio_fanatic : Number
        The ratio of fanatic voters, in the interval [0, 1]. This is used for :meth:`tau`.
    voting_rule : str
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``.

    Notes
    -----
    If the input distribution `d_ranking_share` is not normalized, the profile will be normalized anyway and a
    warning is issued (unless `normalization_warning` is False).

    Examples
    --------
        >>> from fractions import Fraction
        >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
        >>> profile
        ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(3, 5), 'cab': Fraction(3, 10)})
        >>> print(profile)
        <abc: 1/10, bac: 3/5, cab: 3/10> (Condorcet winner: b)
        >>> profile.abc
        Fraction(1, 10)
        >>> profile.d_ranking_share['abc']  # Alternate syntax for profile.abc
        Fraction(1, 10)
        >>> profile.weighted_maj_graph
        array([[0, Fraction(-1, 5), Fraction(2, 5)],
               [Fraction(1, 5), 0, Fraction(2, 5)],
               [Fraction(-2, 5), Fraction(-2, 5), 0]], dtype=object)
        >>> profile.condorcet_winners
        Winners({'b'})
        >>> profile.is_profile_condorcet
        1.0
        >>> profile.has_majority_favorite  # Is one candidate 'top' in a majority of ballots?
        True
        >>> profile.has_majority_ranking  # Does one ranking represent a majority of ballots?
        True
        >>> profile.is_single_peaked  # Is the profile single-peaked?
        True
        >>> profile.support_in_rankings
        {'abc', 'bac', 'cab'}
        >>> profile.is_generic_in_rankings  # Are all rankings there?
        False
    """

    def __init__(self, d_ranking_share, normalization_warning=True, well_informed_voters=True, ratio_fanatic=0,
                 voting_rule=APPROVAL):
        super().__init__(voting_rule=voting_rule)
        # Populate the dictionary
        self._d_ranking_share = DictPrintingInOrderIgnoringZeros({ranking: 0 for ranking in RANKINGS})
        for ranking, share in d_ranking_share.items():
            self._d_ranking_share[ranking] += share
        # Normalize if necessary
        total = sum(self._d_ranking_share.values())
        if not isclose(total, 1.):
            if normalization_warning:
                warnings.warn("Warning: profile is not normalized, I will normalize it.")
            for ranking in self._d_ranking_share.keys():
                self._d_ranking_share[ranking] /= total
        # Other parameters
        self.well_informed_voters = well_informed_voters
        self.ratio_fanatic = ratio_fanatic

    well_informed_voters = property_deleting_cache('_well_informed_voters')
    ratio_fanatic = property_deleting_cache('_ratio_fanatic')

    @cached_property
    def d_ranking_share(self):
        return self._d_ranking_share

    def __repr__(self):
        """
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)},
            ...                          well_informed_voters=False, ratio_fanatic=Fraction(1, 10))
            >>> profile
            ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(3, 5), 'cab': Fraction(3, 10)}, \
well_informed_voters=False, ratio_fanatic=Fraction(1, 10))
        """
        arguments = repr(self._d_ranking_share)
        if not self.well_informed_voters:
            arguments += ', well_informed_voters=False'
        if self.ratio_fanatic > 0:
            arguments += ', ratio_fanatic=%r' % self.ratio_fanatic
        if self.voting_rule != APPROVAL:
            arguments += ', voting_rule=%r' % self.voting_rule
        return 'ProfileOrdinal(%s)' % arguments

    def __str__(self):
        """
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)},
            ...                          well_informed_voters=False, ratio_fanatic=Fraction(1, 10))
            >>> print(profile)
            <abc: 1/10, bac: 3/5, cab: 3/10> (Condorcet winner: b) (badly informed voters) (ratio_fanatic: 1/10)
        """
        result = '<%s>' % str(self._d_ranking_share)[1:-1]
        if self.is_profile_condorcet:
            result += ' (Condorcet winner: %s)' % self.condorcet_winners
        if not self.well_informed_voters:
            result += ' (badly informed voters)'
        if self.ratio_fanatic > 0:
            result += ' (ratio_fanatic: %s)' % self.ratio_fanatic
        if self.voting_rule != APPROVAL:
            result += ' (%s)' % self.voting_rule
        return result

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')

    def __eq__(self, other):
        """Equality test.

        Parameters
        ----------
        other : Object

        Returns
        -------
        bool
            True iff this profile is equal to `other`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> profile == ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            True
        """
        return (isinstance(other, ProfileOrdinal)
                and self.d_ranking_share == other.d_ranking_share
                and self.well_informed_voters == other.well_informed_voters
                and self.ratio_fanatic == other.ratio_fanatic
                and self.voting_rule == other.voting_rule)

    @cached_property
    def standardized_version(self):
        """ProfileOrdinal : Standardized version of the profile (makes it unique, up to permutations of the candidates).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> profile.standardized_version
            ProfileOrdinal({'abc': Fraction(3, 5), 'bac': Fraction(1, 10), 'cba': Fraction(3, 10)})
            >>> profile.is_standardized
            False
        """
        def translate(s, permute):
            return s.replace('a', permute[0]).replace('b', permute[1]).replace('c', permute[2])

        best_d = {}
        best_signature = []
        for perm in XYZ_PERMUTATIONS:
            d_test = {translate(ranking, perm): share for ranking, share in self.d_ranking_share.items()}
            signature_test = [d_test[ranking] for ranking in XYZ_RANKINGS]
            if signature_test > best_signature:
                best_signature = signature_test
                best_d = d_test
        return ProfileOrdinal({ranking: best_d[xyz_ranking] for ranking, xyz_ranking in zip(RANKINGS, XYZ_RANKINGS)},
                              well_informed_voters=self.well_informed_voters, ratio_fanatic=self.ratio_fanatic,
                              voting_rule=self.voting_rule)

    # Has full support
    @cached_property
    def support(self):
        """:class:`SetPrintingInOrder` of str : Alternate notation for :meth:`support_in_rankings`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> profile.support
            {'abc', 'bac', 'cab'}
        """
        return self.support_in_rankings

    @cached_property
    def is_generic(self):
        """bool : Alternate notation for :meth:`is_generic_in_rankings`

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> profile.is_generic
            False
        """
        return self.is_generic_in_rankings

    # Tau and strategy-related stuff

    def tau(self, strategy):
        """Tau-vector associated to a strategy, with partial fanatic voting.

        Parameters
        ----------
        strategy : an argument accepted by :meth:`tau_strategic`

        Returns
        -------
        TauVector
            A share :attr:`ratio_fanatic` of voters vote only for their top candidate, and the rest of the voters
            vote strategically (in the sense of :meth:`tau_strategic`). In other words, this tau-vector
            is the barycenter of ``tau_fanatic`` and ``tau_strategic(strategy)``, with respective
            weights ``self.ratio_fanatic`` and ``1 - self.ratio_fanatic``.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> strategy = StrategyOrdinal({'abc': 'a', 'bac': 'ab', 'cab': 'c'})
            >>> tau = profile.tau(strategy)
            >>> print(tau)
            <a: 1/10, ab: 3/5, c: 3/10> ==> a
            >>> τ = profile.τ(strategy)  # Alternate syntax
            >>> print(τ)
            <a: 1/10, ab: 3/5, c: 3/10> ==> a
        """
        tau_fanatic = self.tau_fanatic
        tau_strategic = self.tau_strategic(strategy)
        t = {ballot: barycenter(a=tau_strategic.d_ballot_share[ballot],
                                b=tau_fanatic.d_ballot_share[ballot],
                                ratio_b=self.ratio_fanatic)
             for ballot in BALLOTS_WITHOUT_INVERSIONS}
        return TauVector(t, voting_rule=self.voting_rule)

    @cached_property
    def tau_fanatic(self):
        """Tau-vector associated to fanatic voting.

        Returns
        -------
        TauVector
            In Approval or Plurality, all voters approve of their top candidate only. In Anti-Plurality, they all
            disapprove of their bottom candidate, i.e. they approve their two first candidates.
        """
        t = {ballot: 0 for ballot in BALLOTS_WITHOUT_INVERSIONS}
        if self.voting_rule in {APPROVAL, PLURALITY}:
            for ranking, share in self.d_ranking_share.items():
                t[ballot_one(ranking)] += share
        elif self.voting_rule == ANTI_PLURALITY:
            for ranking, share in self.d_ranking_share.items():
                t[ballot_one_two(ranking)] += share
        else:
            raise NotImplementedError
        return TauVector(t, voting_rule=self.voting_rule)

    def tau_strategic(self, strategy):
        """Tau-vector associated to a strategy.

        Parameters
        ----------
        strategy : StrategyOrdinal
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        TauVector
            Tau-vector associated to this profile and strategy `strategy`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> strategy = StrategyOrdinal({'abc': 'a', 'bac': 'ab', 'cab': 'c'})
            >>> tau_strategic = profile.tau_strategic(strategy)
            >>> print(tau_strategic)
            <a: 1/10, ab: 3/5, c: 3/10> ==> a

        In the case of approval with badly informed voters, we do as if the tau-vector were the vector of scores (up
        to a renormalization):

            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)},
            ...                          well_informed_voters=False)
            >>> strategy = StrategyOrdinal({'abc': 'a', 'bac': 'ab', 'cab': 'c'})
            >>> tau_strategic = profile.tau_strategic(strategy)
            >>> print(tau_strategic)
            <a: 7/16, b: 3/8, c: 3/16> ==> a
        """
        assert self.voting_rule == strategy.voting_rule
        t = {ballot: 0 for ballot in BALLOTS_WITHOUT_INVERSIONS}
        for ranking, ballot in strategy.d_ranking_ballot.items():
            if self.d_ranking_share[ranking] > 0:
                t[ballot] += self.d_ranking_share[ranking]
        if self.voting_rule == APPROVAL and not self.well_informed_voters:
            return TauVector(
                {'a': t['a'] + t['ab'] + t['ac'], 'b': t['b'] + t['ab'] + t['bc'], 'c': t['c'] + t['ac'] + t['bc']},
                normalization_warning=False, voting_rule=self.voting_rule)
        else:
            return TauVector(t, voting_rule=self.voting_rule)

    def is_equilibrium(self, strategy):
        """Whether a strategy is an equilibrium.

        Parameters
        ----------
        strategy : StrategyOrdinal
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        EquilibriumStatus
            Whether `strategy` is an equilibrium in this profile.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> strategy = StrategyOrdinal({'abc': 'a', 'bac': 'ab', 'cab': 'c'})
            >>> profile.is_equilibrium(strategy)
            EquilibriumStatus.EQUILIBRIUM
        """
        d_ranking_best_response = self.tau(strategy).d_ranking_best_response
        status = EquilibriumStatus.EQUILIBRIUM
        for ranking, share in self.d_ranking_share.items():
            if share == 0:
                continue
            best_response = d_ranking_best_response[ranking]
            if best_response.ballot == INCONCLUSIVE:
                status = min(status, EquilibriumStatus.INCONCLUSIVE)  # pragma: no cover
            elif best_response.ballot == UTILITY_DEPENDENT:
                status = min(status, EquilibriumStatus.UTILITY_DEPENDENT)
            elif strategy.d_ranking_ballot[ranking] != best_response.ballot:
                return EquilibriumStatus.NOT_EQUILIBRIUM
        return status

    @cached_property
    def analyzed_strategies(self):
        """AnalyzedStrategies : Analyzed strategies of the profile.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> profile.analyzed_strategies
            Equilibria:
            <abc: a, bac: b, cab: ac> ==> b (FF)
            <abc: a, bac: ab, cab: c> ==> a (D)
            <BLANKLINE>
            Utility-dependent equilibrium:
            <abc: ab, bac: b, cab: c> ==> b (FF)
            <BLANKLINE>
            Non-equilibria:
            <abc: a, bac: b, cab: c> ==> b (FF)
            <abc: a, bac: ab, cab: ac> ==> a (D)
            <abc: ab, bac: b, cab: ac> ==> b (FF)
            <abc: ab, bac: ab, cab: c> ==> a, b (FF)
            <abc: ab, bac: ab, cab: ac> ==> a (D)
            >>> print(profile.analyzed_strategies.equilibria[0])
            <abc: a, bac: b, cab: ac> ==> b
        """
        equilibria = []
        utility_dependent = []
        inconclusive = []
        non_equilibria = []
        d_ranking_possible_ballots = {ranking: [ballot_low_u(ranking, self.voting_rule),
                                                ballot_high_u(ranking, self.voting_rule)]
                                      if self.d_ranking_share[ranking] > 0 else ['']
                                      for ranking in RANKINGS}
        for d_ranking_ballot in product_dict(d_ranking_possible_ballots):
            strategy = StrategyOrdinal(d_ranking_ballot, profile=self, voting_rule=self.voting_rule)
            status = strategy.is_equilibrium
            if status == EquilibriumStatus.EQUILIBRIUM:
                equilibria.append(strategy)
            elif status == EquilibriumStatus.UTILITY_DEPENDENT:
                utility_dependent.append(strategy)
            elif status == EquilibriumStatus.INCONCLUSIVE:  # pragma: no cover
                inconclusive.append(strategy)
                warnings.warn('Met an inconclusive case: \nprofile = %r\nstrategy = %r'
                              % (self, strategy))
            else:
                non_equilibria.append(strategy)
        return AnalyzedStrategies(equilibria, utility_dependent, inconclusive, non_equilibria)

    def proba_equilibrium(self, test=None):
        """Probability that an equilibrium exists (depending on the utilities).

        Parameters
        ----------
        test : callable
            A function ``StrategyOrdinal -> bool`` that gives a condition on the strategy. Default: always True.

        Returns
        -------
        float
            The probability that an equilibrium strategy exists, that meets the `test` condition.

        Notes
        -----
        The result is exact (not based on a Monte-Carlo estimation).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> profile.proba_equilibrium()
            1
        """
        if test is None:
            # noinspection PyUnusedLocal
            def test(strategy):
                return True
        if any([test(strategy) for strategy in self.analyzed_strategies.equilibria]):
            return 1
        support = sorted(self.support)
        dim = len(support)
        masks = [
            [(strategy.d_ranking_best_response[ranking].threshold_utility, len(strategy.d_ranking_ballot[ranking]) == 2)
             for ranking in support]
            for strategy in self.analyzed_strategies.utility_dependent if test(strategy)
        ]
        return masks_area(inf=[0] * dim, sup=[1] * dim, masks=masks)

    def distribution_equilibria(self, test=None):
        """Distribution of numbers of equilibria (depending on the utilities).

        Parameters
        ----------
        test : callable
            A function ``StrategyOrdinal -> bool`` that gives a condition on the strategy. Default: always True.

        Returns
        -------
        list
            A list that represents an histogram. The distribution of number of equilibria (meeting the `test` condition).

        Notes
        -----
        The result is exact (not based on a Monte-Carlo estimation).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> profile.distribution_equilibria()
            array([0.        , 0.        , 0.86290531, 0.13709469])
        """
        if test is None:
            # noinspection PyUnusedLocal
            def test(strategy):
                return True
        cover_alls = np.sum([test(strategy) for strategy in self.analyzed_strategies.equilibria], dtype=int)
        support = sorted(self.support)
        dim = len(support)
        masks = [
            [(strategy.d_ranking_best_response[ranking].threshold_utility, len(strategy.d_ranking_ballot[ranking]) == 2)
             for ranking in support]
            for strategy in self.analyzed_strategies.utility_dependent if test(strategy)
        ]
        return masks_distribution(inf=np.zeros(dim), sup=np.ones(dim), masks=masks, cover_alls=cover_alls)

    def distribution_winners(self, test=None):
        """Distribution of the number of equilibrium winners (depending on the utilities).

        Parameters
        ----------
        test : callable
            A function ``StrategyOrdinal -> bool`` that gives a condition on the strategy. Default: always True.

        Returns
        -------
        list
            A list that represents an histogram. The distribution of number of possible equilibrium winner (with
            strategies that meet the `test` condition).

        Notes
        -----
        The result is exact (not based on a Monte-Carlo estimation).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileOrdinal({'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)})
            >>> profile.distribution_winners()
            array([0., 0., 1., 0.])
        """
        if test is None:
            # noinspection PyUnusedLocal
            def test(strategy):
                return True
        cover_alls = set.union(*([set()] + [
            strategy.winners for strategy in self.analyzed_strategies.equilibria
            if test(strategy)
        ]))
        support = sorted(self.support)
        dim = len(support)
        masks_winners = [
            (
                [(strategy.d_ranking_best_response[ranking].threshold_utility,
                  len(strategy.d_ranking_ballot[ranking]) == 2)
                 for ranking in support],
                strategy.winners
            )
            for strategy in self.analyzed_strategies.utility_dependent if test(strategy)
        ]
        return winners_distribution(inf=np.zeros(dim), sup=np.ones(dim), masks_winners=masks_winners,
                                    cover_alls=cover_alls)
